/***************************************************************/
/*							       */
/*			 POLKAWINDOW.C    		       */
/*							       */
/***************************************************************/
/*    Copyright 1993 Georgia Tech -- John T. Stasko            */

#include "polka_local.H"


static int askingForInput = 0;
static int gotPoint = 0;
static int xPoint,yPoint;


View::View()
{
   debug = 0;
   created = 0;
   time = 0;
   window = 0;   /* need it this way to start (no resize) */
}



int
View::Create(char *title)
{
   Window w;

   PolkaInit();   // just to be safe

   if (created) {
      cerr << "Cannot Create a View already created" << endl;
      return(0);
      }
   else
      created = 1;

   setupWindow(title);  /* sets up X window and Xt widgets */

   /* Do default values */
   pixX = 0;	      /* Width & height init'd in anim_setup */
   pixY = 0;	      /*    (user may resize window)         */
   olx = lx = 0.0;
   orx = rx = 1.0;
   oty = ty = 1.0;
   oby = by = 0.0;
   dx = dy = 1.0;
   bgcolor = WHITE;
   numobjects = 0;
   motionblur = 0;
   AOHead = AOTail = NULL;
   PendingHead = PendingTail = NULL;

   w = XtWindow(easel);
   prepareAnim(w);  /* sets destwin & window member variables */

   fixPad();   // calculates boxpad 

   /* Wait until widgets are displayed */
//   XtAppNextEvent(_context,&xev);
//   XtDispatchEvent(&xev);
//   CheckXEvents();
//   XSync(_display,1);
   return(1);
}



int
View::CheckInput()
{ 
   CheckXEvents(); 
   return(0);
}


void
View::SetBgColor(char *col)
{
   COLORINDEX newcol;
   newcol = load_color(col);
   if (newcol != bgcolor) {
      bgcolor = newcol;
      Refresh();
      }
}




void
View::Refresh()
{
   clear();
   damlx = lx;
   damrx = rx;
   damby = by;
   damty = ty;
   drawAll();
   nextFrame();
   XFlush(_display);
}
   



void
View::clear()
{
   XRectangle rect;
   if (debug)
      printf("DEBUG: View::clear()\n");

   int oldcolor = set_color(bgcolor);

   XSetFillStyle(_display, inq_gc(), FillSolid);
   rect.x = 0;
   rect.y = 0;         // must reset gc's clip area to whole window
   rect.width = pixX;
   rect.height = pixY;
   XSetClipRectangles(_display,inq_gc(),0,0,&rect,1,YXBanded);
   XFillRectangle(_display, window, inq_gc(), 0, 0, pixX, pixY);
   set_color(oldcolor);
}


void
View::originateAnimObjects(int frame)
{
   AOPtr aop,old;

   for (aop=PendingHead; aop; ) 
      if (aop->object->OriginateTime() <= frame) {
         // remove from Pending list
         if (aop == PendingHead)
            PendingHead = aop->next;
         else
            aop->prev->next = aop->next;
         if (aop == PendingTail)
            PendingTail = aop->prev;
         else
            aop->next->prev = aop->prev;
         old = aop;
         aop = aop->next;

         // add to originated list
         if (!AOHead)
            AOTail = old;
         else
            AOHead->prev = old;
         old->next = AOHead;
         old->prev = NULL;
         AOHead = old;

         old->object->SetOriginated(1);
         old->object->DamageIt();
         }
      else
        aop = aop->next;
}



void
View::updateAll(int frame)
{
   AOPtr head,walker,curr;

   if (debug)
      printf("DEBUG: View::UpdateAll %d\n",frame);

   head = AOHead;      // Have to be careful here.  Delete ops and
   walker = AOTail;    // Raise and Lower actions reorder the AO List.
   curr = NULL;
   while (curr != head) {
      curr = walker;
      walker = walker->prev;
      curr->object->Update(frame);
      }
}


void
View::drawAll()
{
   AOPtr aop;
   double damagelx,damagerx,damageby,damagety;

   if (debug)
      printf("DEBUG: View::checkDrawAll\n");

   if (window) {
      damagelx = damlx-boxpad;
      damagerx = damrx+boxpad;
      damageby = damby-boxpad;
      damagety = damty+boxpad;
      for (aop=AOTail; aop; aop=aop->prev)
          if (aop->object->Intercepts(damagelx,damageby,damagerx,damagety))
             aop->object->Draw();
      }
}


void
View::nextFrame()
{
   if (debug)
      printf("DEBUG: View::NextFrame\n");

//   batchMode(1);
   animNextFrame();
//   batchMode(0);
}



int
View::Animate(int start, int frames)
{
   int t,cliplx,clipby,cliprx,clipty;
   XRectangle xrect;

   
   if (debug)
      printf("DEBUG: View::Animate start=%d for %d frames\n",start,frames);

   BatchMode(1);
   for (t=0; t<frames; t++, start++)
       { CheckXEvents();

         damlx = rx; damrx = lx;    // init damaged area
         damby = ty; damty = by;

         originateAnimObjects(start);
         updateAll(start);     // make changes for this frame

         if (damrx > damlx) {     // somebody did something
            // printf("Accum (lf)  (%.3lf, %.3lf)   (%.3lf, %.3lf)\n",
            //        damlx,damby,damrx,damty);
            cliplx = TO_PIX_X(damlx) - 1;
            cliprx = TO_PIX_X(damrx) + 1;
            clipby = TO_PIX_Y(damby) + 1;
            clipty = TO_PIX_Y(damty) - 1;
            if (cliplx < 0) cliplx=0;
            if (clipty < 0) clipty=0;
            if (cliprx > pixX) cliprx=pixX;
            if (clipby > pixY) clipby=pixY;
            // printf("Accum (int)  (%d, %d)   (%d, %d)\n",
            //          cliplx,clipby,cliprx,clipty);

            xrect.x = cliplx;   // pad these guys to be safe
            xrect.y = clipty;
            xrect.width = cliprx-cliplx;
            xrect.height = clipby-clipty;
            XSetClipRectangles(_display,inq_gc(),0,0,&xrect,1,YXBanded);

            int oldcolor = set_color(bgcolor);
            XSetFillStyle(_display, inq_gc(), FillSolid);
            XFillRectangle(_display, window, inq_gc(), cliplx, clipty, 
                            cliprx, clipby);
            set_color(oldcolor);
         
            drawAll();  // redraw offscreen only if in damage area
            }

         nextFrame();          // blt onscreen
         if (_delay)
            micro_sleep(_delay);
       }
   BatchMode(0);

   return(start);
}



int
View::AnimateOne(int start)
{
   int t,cliplx,clipby,cliprx,clipty;
   XRectangle xrect;
   
   if (debug)
      printf("DEBUG: View::animate1 start=%d\n",start);

   damlx = rx; damrx = lx;    // init damaged area
   damby = ty; damty = by;

   originateAnimObjects(start);
   updateAll(start);     // make changes for this frame

   if (damrx > damlx) {     // somebody did something
      // printf("Accum (lf)  (%.3lf, %.3lf)   (%.3lf, %.3lf)\n",
      //        damlx,damby,damrx,damty);
      cliplx = TO_PIX_X(damlx) - 1;
      cliprx = TO_PIX_X(damrx) + 1;
      clipby = TO_PIX_Y(damby) + 1;
      clipty = TO_PIX_Y(damty) - 1;
      if (cliplx < 0) cliplx=0;
      if (clipty < 0) clipty=0;
      if (cliprx > pixX) cliprx=pixX;
      if (clipby > pixY) clipby=pixY;
      // printf("Accum (int)  (%d, %d)   (%d, %d)\n",
      //          cliplx,clipby,cliprx,clipty);

      xrect.x = cliplx;   // pad these guys to be safe
      xrect.y = clipty;
      xrect.width = cliprx-cliplx;
      xrect.height = clipby-clipty;
      XSetClipRectangles(_display,inq_gc(),0,0,&xrect,1,YXBanded);

      int oldcolor = set_color(bgcolor);
      XSetFillStyle(_display, inq_gc(), FillSolid);
      XFillRectangle(_display, window, inq_gc(), cliplx, clipty, 
                      cliprx, clipby);
      set_color(oldcolor);
         
      drawAll();  // redraw offscreen only if in damage area
      }

   nextFrame();          // blt onscreen
   return(start+1);
}





int
View::PickCoord(double& x, double& y)
{
   XEvent event;

   askingForInput = 1;
   /* Confine mouse pointer to main (easel) window */
   XGrabButton(_display, AnyButton, AnyModifier, destwin, TRUE,
	       ButtonPressMask | ButtonMotionMask | ButtonReleaseMask,
	       GrabModeAsync, GrabModeAsync, destwin,
	       XCreateFontCursor(_display, XC_crosshair));

   /* Wait for button press event handler to change "getMouse" */
   while (!gotPoint) {
      XtAppNextEvent(_context,&event);
      XtDispatchEvent(&event);
      }

   XUngrabButton(_display, AnyButton, AnyModifier, destwin);

   x = TO_REAL_X(xPoint);
   y = TO_REAL_Y(yPoint);

   gotPoint = 0;
   askingForInput = 0;
   return(1);
}



int
View::PickAnimObject(AnimObject* &ao)
{
   XEvent event;
   double xpt, ypt;
   AOPtr  list;
   double bblx,bbby,bbrx,bbty;
   double delta=0.015;

   askingForInput = 1;
   /* Confine mouse pointer to main (easel) window */
   XGrabButton(_display, AnyButton, AnyModifier, destwin, TRUE,
	       ButtonPressMask | ButtonMotionMask | ButtonReleaseMask,
	       GrabModeAsync, GrabModeAsync, destwin,
	       XCreateFontCursor(_display, XC_crosshair));

   /* Wait for button press event handler to change "getMouse" */
   while (!gotPoint) {
      XtAppNextEvent(_context,&event);
      XtDispatchEvent(&event);
      }

   XUngrabButton(_display, AnyButton, AnyModifier, destwin);

   xpt = TO_REAL_X(xPoint);
   ypt = TO_REAL_Y(yPoint);

   ao = NULL;
   for (list=AOTail; list; list=list->prev) {
      list->object->BoundBox(&bblx, &bbby, &bbrx, &bbty);
      // deltas are for things like horiz and vert lines
      if ((bblx-delta <= xpt) && (xpt <= bbrx+delta) 
          && (bbby-delta <= ypt) && (ypt <= bbty+delta))
         ao = list->object;
      }
   gotPoint = 0;
   return( (ao ? 1 : 0));
}


void
SelectPoint(Widget, XtPointer xtp, XEvent *event, Boolean *)
{
   double xpt, ypt;
   AnimObject *ao;
   AOPtr  list;
   double bblx,bbby,bbrx,bbty;
   double delta=0.015;
   View *v = (View*)xtp;

   if (askingForInput) { // user has explicitly requested a pick
      gotPoint = 1;
      xPoint = event->xbutton.x;
      yPoint = event->xbutton.y;
      }
   else {   // random pick, so check for object callback
      xpt = v->TO_REAL_X(event->xbutton.x);
      ypt = v->TO_REAL_Y(event->xbutton.y);

      ao = NULL;
      for (list=v->AOTail; list; list=list->prev) {
         list->object->BoundBox(&bblx, &bbby, &bbrx, &bbty);
         // deltas are for things like horiz and vert lines
         if ((bblx-delta <= xpt) && (xpt <= bbrx+delta) 
                    && (bbby-delta <= ypt) && (ypt <= bbty+delta)) {
            ao = list->object;
            break;
            }
         }
      if (ao && ao->callbackFct) 
         ao->callbackFct(ao, ao->callbackData, xpt, ypt); 
      }
}









