/***************************************************************/
/*							       */
/*			  POLKA_LOCAL.H			       */
/*							       */
/***************************************************************/
/*    Copyright 1993 Georgia Tech -- John T. Stasko            */


/***************************************************************/
/**********	    include files		      **********/
/***************************************************************/


/* #include <stream.h>  */
/* #include <string.h> */
/* #include <libc.h> */

#include <stdlib.h>
#include <iostream.h>
#include <stdio.h>
/* #include <strings.h> */
#include <stdarg.h>
#include <limits.h>
#include <math.h>
#include "polka.H"

/***************************************************************/
/**********	  constant definitions		      **********/
/***************************************************************/

#define   WIDTH             500
#define   HEIGHT            500
#define   MAXSTRINGLENGTH   256
#define   ZERO              0
#define   MYPI              ((double)3.14159)

#define WHITE             0
#define YELLOW            1
#define GREEN             2
#define BLUE              3
#define ORANGE            4
#define RED               5
#define MAROON            6
#define BLACK             7

#define   POLKA_BASE_FILL         0
#define   POLKA_NUM_FILL         41
#define   POLKA_FILL_OUTLINE     POLKA_BASE_FILL
#define   POLKA_FILL_SOLID       POLKA_BASE_FILL + POLKA_NUM_FILL - 1

#define   POLKA_ACTION_POINTS      20
#define   POLKA_ACTION_NUMTYPES    3


#define   MIN(A,B)        ( (A) < (B) ? (A) : (B) )

#ifndef ABS
#define  ABS(A)       ( (A) < 0.0 ? -(A) : (A) )
#endif

#define  BETWEEN(A,X,B) ( ((A) <= (X)) && ((X) <= (B)) )

#define POLKA_complain(msg1,msg2) (fprintf(stderr, "Error: %s: %s.\n", (msg1), (msg2)))

/***************************************************************/
/**********	      local data types		      **********/
/***************************************************************/

typedef struct TimerCBNode
{
   int                 time;
   TIMER_CALLBACK      callbackFct;
   void               *data;
   struct TimerCBNode *next;
} *TimerCBPtr;


typedef struct ImplNode
{
   class AnimObjectImpl *objimpl;
   int                   time;
   int                   existingpos;
   struct ImplNode      *next;
} *ImplPtr;

typedef struct IntNode
{
   int             value;
   struct IntNode *next;
} *IntPtr;


typedef struct ActionDeleteNode
{
   int             value;
   struct Action  *act;
   struct ActionDeleteNode *next;
} *ActionDeletePtr;


typedef struct ActionInfo
{
   int                time;
   struct OFFSET     *offsets;        // always points to start of list
   struct OFFSET     *doing;          // the one currently at
   struct Action     *originalAction; // used for deleting Continuous Actions
   struct Action     *action;         // a copy of the Action being Programmed
   struct ActionInfo *next;
   struct ActionInfo *prev;
 } *ActionInfoPtr;

typedef struct OFFSET        /* used by Actions */
{
   double	  dx;
   double	  dy;
   struct OFFSET *prevo;
   struct OFFSET *nexto;
} *OFFSET_PTR;



typedef struct _COLOR_MAPPER {
   char        name[30];
   COLORINDEX  colornum;
   double      pathval;
   struct _COLOR_MAPPER *next;
   } COLOR_MAPPER;



/***************************************************************/
/**********	    global variables		      **********/
/***************************************************************/

extern   int           _delay;
extern   Display      *_display;
extern   XtAppContext  _context;
extern   int           _color_screen;
extern   COLOR_MAPPER *_color_list;

void              PolkaInit();
void              BatchMode(int);
void              CheckXEvents();
void              micro_sleep(int);


void              SelectPoint(Widget, XtPointer, XEvent *, Boolean *);

/******  In polkawidget.C   ******/
void              resize(Widget, XtPointer, XtPointer); 


/******  In polkagc.C  ********/
void              SetBackground(COLORINDEX);
COLORINDEX        load_color(COLOR);
Pixel             GetXColor(COLORINDEX);
COLORINDEX        set_color(COLORINDEX);
LINE_STYLE        line_style(LINE_STYLE);
FILL_STYLE        fill_style(FILL_STYLE);
int               load_font(char *);
void              text_info(Font, char *, int *, int *, int *, int *);
int               set_font(int);
GC                inq_gc();
int               inq_color();
int               inq_font();
void              initGC();


COLORINDEX        get_pathcolor(double, double);
COLORINDEX        get_color(COLOR);
FILL_STYLE        get_fillstyle(double);
LINE_STYLE        get_linestyle(double, double);
void              compute_arrow_off(double, double, double *, double *, 
                                     double *, double *);
void              arrow_poly_direction(int, double[7], double[7], 
                                       double *, double *, double *, double *);

/******* In polkaspline.C *******/
void              make_a_spline(int, int[], int[], View *);

/******** In polkaaux.C ********/
int               streql(char *, char *);
int               streqln(char *, char *, int);
int               streql(char *, const char *);
int               streqln(char *, const char *, int);
int               streql(const char *, const char *);
int               streqln(const char *, const char *, int);

/******** In polkagroup.C ********/
int lineArray(View *, AnimObject *[], int, double, double, double, 
            double, AnimObjectGroup *);
int rectangleArray(View *, AnimObject *[], int, double, double, double, 
            double, AnimObjectGroup *);
int circleArray(View *, AnimObject *[], int, double, double, double, 
            double, AnimObjectGroup *);

