/***************************************************************/
/*							       */
/*			     POLKAOBJECT.C		       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

#include "polka_local.H"

AnimObject::AnimObject(AnimObjectImpl *i) : object(i)
{
   type = i->Type();
   view = i->MyView();
   first = last = NULL;
   firstContinuous = lastContinuous = NULL;
   deleters = NULL;
   changers = NULL;
   continDeleters = NULL;
   attachments = NULL;
   attachedTo = NULL;
   attachDoneFrom = NULL;
   originateAt = INT_MAX;   // from math.h
   originated = 0;
   modified = 1;
   updating = 1;
   callbackFct = NULL;
   callbackData = NULL;
   objs = NULL;
   i->SetWrapper(this);
   addObject();
}



AnimObject::AnimObject(const AnimObject& a)
{
   AnimObjectImpl *ooo;

   ooo = a.object;   // this is a kludge for a CC compiler bug
   object = ooo->clone();

   type = a.type;
   view = a.view;
   modified = 1;
   updating = 1;
   firstContinuous = lastContinuous = NULL;
   first = last = NULL;
   deleters = NULL;
   changers = NULL;
   continDeleters = NULL;
   attachments = NULL;
   attachedTo = NULL;
   attachDoneFrom = NULL;
   originated = 0;
   originateAt = INT_MAX;   // from math.h
   callbackFct = a.callbackFct;
   callbackData = a.callbackData;
   objs = NULL;
   object->SetWrapper(this);
   addObject();
}



AnimObject&
AnimObject::operator=(const AnimObject& a)
{
   AnimObjectImpl *ooo;

   delete object;    // out with the old (from bitwise construction)

   ooo = a.object;   // this is a kludge for a CC compiler bug
   object = ooo->clone();

   type = a.type;
   view = a.view;
   modified = 1;
   updating = 1;
   firstContinuous = lastContinuous = NULL;
   first = last = NULL;
   deleters = NULL;
   changers = NULL;
   continDeleters = NULL;
   attachments = NULL;
   attachedTo = NULL;
   attachDoneFrom = NULL;
   originated = 0;
   originateAt = INT_MAX;   // from math.h
   callbackFct = a.callbackFct;
   callbackData = a.callbackData;
   objs = NULL;
   object->SetWrapper(this);
   addObject();
   return *this;
}




void
AnimObject::addObject()
{
   if (type == P_Set) return;  // did this rather than making a virtual
                               // Sets never go in AnimObject list
   AOPtr aop = new AONode;
   aop->object = this;
   if (!view->PendingHead)
      view->PendingHead = aop;
   else
      view->PendingTail->next = aop;
   aop->prev = view->PendingTail;
   aop->next = NULL;
   view->PendingTail = aop;

   viewAOP = aop;
}



void
AnimObject::deleteMe()  // this should never be called on a Set
{
   ActionInfoPtr a,old;
   ActionDeletePtr adp,adpold;
   ImplPtr i,iold;
   IntPtr t,told;
   AOPtr aop,before;
   Attachment att,attold;

   a = first;
   while (a) {   // get rid of programming nodes
       old = a;
       a = a->next;
       delete(old->action);
       delete old;
   }

   a = firstContinuous;
   while (a) {   // get rid of continuous programming nodes
       old = a;
       a = a->next;
       delete(old->action);
       delete old;
   }

   i = changers;
   while (i) {   // get rid of change nodes
      iold = i;
      i = i->next;
      delete(iold);
   }

   t = deleters;
   while (t) {   // get rid of delete nodes
      told = t;
      t = t->next;
      delete(told);
   }

   adp = continDeleters;
   while (adp) {   // get rid of delete nodes
      adpold = adp;
      adp = adp->next;
      delete(adpold);
   }

   att = attachments;
   while (att) {   // get rid of attachment nodes
      attold = att;
      att = att->next;
      attold->object->attachedTo = NULL;  // remove reference to this one
      delete(attold);
   }

   for (aop=objs; aop; ) {  // remove object from any sets it was in
      ((Set *)(aop->object))->RemoveFromSet(this);
      before = aop;
      aop = aop->next;
      delete(before);
    }

   if (attachedTo)   // make sure others pointing to this disregard it
      attachedTo->removeAttachment(this);

   if (originated) {
      for (aop=view->AOHead; aop; aop=aop->next) {  
        if (aop->object == this) {
           if (aop->next)
              aop->next->prev = aop->prev;
           if (aop->prev)
              aop->prev->next = aop->next;
           if (aop == view->AOHead)
              view->AOHead = aop->next;
           if (aop == view->AOTail)
              view->AOTail = aop->prev;
           delete(aop);
           return;
         }
      }
   }
   else { // not originated
      for (aop=view->PendingHead; aop; aop=aop->next) {  
        if (aop->object == this) {
           if (aop->next)
              aop->next->prev = aop->prev;
           if (aop->prev)
              aop->prev->next = aop->next;
           if (aop == view->PendingHead)
              view->PendingHead = aop->next;
           if (aop == view->PendingTail)
              view->PendingTail = aop->prev;
           delete(aop);
           return;
         }
      }
   }
}




void
AnimObject::Originate(int ti)
{
   if (originated) {
      cout << "Attempt to Originate AnimObject already Originated" << endl;
      return;
      }
   originateAt = ti;
}

   


void
AnimObject::Change(int ti, AnimObject *newobj, int useoldpos)
{
   ImplPtr run;   // add a change node, keep them in numerical order

   if ((type == P_Set) || (newobj->Type() == P_Set)) {
      cout << "Changing to/from a Set AnimObject is not allowed" << endl;
      return;
    }

   if (updating == 0)    // not on Update list (which is 1 or -1)
      view->AddUpdateObject(this->viewAOP);
   updating = 1;

   ImplPtr ip = new ImplNode;
   ip->time = ti;
   ip->existingpos = useoldpos;
   ip->objimpl = (newobj->Object())->clone();
   if ((!changers) || (changers->time >= ti)) {
      ip->next = changers;
      changers = ip;
      }
   else {
      run = changers;
      while ((run->next) && (run->next->time < ti)) 
         run = run->next;
      ip->next = run->next;
      run->next = ip;
      }
}



void
AnimObject::Delete(int ti)
{
   IntPtr run;   // add a delete node, keep them in numerical order

   if (type == P_Set) // since Sets are in AnimObj list, makes no sense
      return;

   if (updating == 0)     // not on Update list (which is 1 or -1)
      view->AddUpdateObject(this->viewAOP);
   updating = 1;

   IntPtr ip = new IntNode;
   ip->value = ti;
   if ((!deleters) || (deleters->value >= ti)) {
      ip->next = deleters;
      deleters = ip;
      }
   else {
      run = deleters;
      while ((run->next) && (run->next->value < ti)) 
         run = run->next;
      ip->next = run->next;
      run->next = ip;
      }
}




void
AnimObject::RemoveContinuous(int ti, ActionPtr ap)
{
   ActionDeletePtr run;   // add a deletion node, keep them in numerical order

   if (type == P_Set) {   // sets are special
      for (AOPtr a=objs; a; a=a->next)
         a->object->RemoveContinuous(ti,ap);
      return;
   }

   ActionDeletePtr adp = new ActionDeleteNode;
   adp->value = ti;
   adp->act = ap;
   if ((!continDeleters) || (continDeleters->value >= ti)) {
      adp->next = continDeleters;
      continDeleters = adp;
      }
   else {
      run = continDeleters;
      while ((run->next) && (run->next->value < ti)) 
         run = run->next;
      adp->next = run->next;
      run->next = adp;
      }
}




int
AnimObject::Program(int t, ActionPtr ap)
{
   int r = 0;

   if (view->debug) 
      printf("DEBUG: AnimObject::Program at frame=%d\n",t);

   if (type == P_Set) {   // sets are special
      for (AOPtr a=objs; a; a=a->next)
         r = a->object->Program(t,ap);
      return(r);
   }

   if (updating == 0)    // not on Update list (which is 1 or -1)
      view->AddUpdateObject(this->viewAOP);
   updating = 1;

   ActionInfoPtr aip = new ActionInfo;
   aip->time = t;
   aip->originalAction = ap;
   aip->action = ap->Copy();
   aip->offsets = aip->doing = aip->action->head;
   aip->next = NULL;
   aip->prev = last;
   if (!first)
      { first = last = aip;
      }
   else
      { last->next = aip;
        last = aip;
      }
   return(ap->count);
}





int
AnimObject::ProgramContinuous(int t, ActionPtr ap)
{
   int r = 0;

   if (view->debug) 
      printf("DEBUG: AnimObject::ProgramContinuous at frame=%d\n",t);

   if (type == P_Set) {   // sets are special
      for (AOPtr a=objs; a; a=a->next)
         r = a->object->ProgramContinuous(t,ap);
      return(r);
   }

   if (updating == 0)    // not on Update list (which is 1 or -1)
      view->AddUpdateObject(this->viewAOP);
   updating = 1;

   ActionInfoPtr aip = new ActionInfo;
   aip->time = t;
   aip->originalAction = ap;
   aip->action = ap->Copy();
   aip->offsets = aip->doing = aip->action->head;
   aip->next = NULL;
   aip->prev = lastContinuous;
   if (!firstContinuous)
      { firstContinuous = lastContinuous = aip;
      }
   else
      { lastContinuous->next = aip;
        lastContinuous = aip;
      }
   return(ap->count);
}





void
AnimObject::Update(int ti)    // should never get called on a Set
{
   ActionInfoPtr temp, old = NULL;
   int finished;
   Attachment att;

   if (view->debug)
      printf("DEBUG: AnimObject::Update time=%d\n",ti);

 /* This is complicated.  Basically, when an AnimObject has no programming*/
 /* anymore, we don't immediately take it off the updating list.  We wait one*/
 /* frame to do that.  This is so that if you're in the mode where only*/
 /* updated objects get redrawn, we make sure to redraw the object the last*/
 /* time it's on the update list.  If we took an object off the list right*/
 /* away, then we walked the update list to redraw, we would miss it.*/
 /* Below we've come around the second time on an object with no programming*/

   if (updating == -1) {  // pending removal from Update list, see below
      view->RemoveUpdateObject(this->viewAOP);
      updating = 0;
      return;
      }

   if (deleters && doDeletes(ti)) 
      return;

   doContDeletes(ti);  // see if contin Actions are scheduled to be removed

   temp = first;
   while (temp)
   { finished = 0;
     if (temp->time < ti) // somehow skipped animating some frames
        { while ((temp->doing) && (temp->time < ti))
             { object->Trans(temp->action->type,
                                 temp->doing->dx,temp->doing->dy);
               temp->doing = temp->doing->nexto;
               if (!(temp->doing)) 
                  { finished = 1;
                    old = temp;
                    break;
                  }
               temp->time++;
  	     }
        }
     if (!finished)
        { if (temp->time == ti)            /* perform the action */
             { object->Trans(temp->action->type,
                               temp->doing->dx,temp->doing->dy);
               temp->doing = temp->doing->nexto;
               temp->time++;
               if (!(temp->doing))
	          { old = temp;
                    finished = 1;
  		  }
               else
                  temp = temp->next;
	     }
          else
             temp = temp->next;
        }
     if (finished)
        temp = DeleteActionInfoPtr(old);
   }  // while

   temp = firstContinuous;
   while (temp) {
      if ((temp->time <= ti) && (temp->doing)) {
         object->Trans(temp->action->type,
                               temp->doing->dx,temp->doing->dy);
         temp->doing = temp->doing->nexto;
         if (!(temp->doing))   // hit end of list, so go back
            temp->doing = temp->offsets;
         }
      temp = temp->next;
      }
            
   doChanges(ti);

   attachDoneFrom = this;
   for (att=attachments; att; att=att->next)
       att->object->propagateAttachment(this,att,this);
   attachDoneFrom = NULL;

   if (!first && !firstContinuous && !changers && !deleters) 
      updating = -1;
        // nothing is pending anymore, get ready to move back to plain list
        // It will happen next time around
        // Don't want to remove it here because we want it to get
        //     drawn by drawAllUpdaters
}

              


Attachment
AnimObject::AttachTo(AnimObject *ao, PART dept, PART indept,
                     double xoff, double yoff)
{
   Attachment a;

   if (type == P_Set) {
      cout << "Cannot attach a Set to another AnimObject" << endl;
      return NULL;
    }
   if (ao->Type() == P_Set) {
      cout << "Attaching to a Set AnimObject is not allowed" << endl;
      return NULL;
    }

   // check if ao already has attachment, if so, remove it
   if (attachedTo)
      ao->removeAttachment(this);

   a = new AttachNode;
   a->object = this;
   a->indept = indept;
   a->dept = dept;
   a->xoffset = xoff;
   a->yoffset = yoff;
   a->next = NULL;
   ao->addAttachment(a);
   attachedTo = ao;
   return(a);
}



int
AnimObject::Detach(Attachment att)
{
   Attachment a;

   if (att->object != this)
      return(0);
   if (!attachedTo) 
      return(0);
   for (a=attachedTo->getAttachments(); a; a=a->next)
       if (a == att) break;
   if (!a)
      return(0);

   attachedTo->removeAttachment(this);
   attachedTo = NULL;
   delete(att);
   return(1);
}


void
AnimObject::addAttachment(Attachment a)
{
   a->next = attachments;
   attachments = a;
}


void
AnimObject::removeAttachment(AnimObject *remover)
{
   Attachment att,old;

   if (!attachments) return;

   if (attachments->object == remover) {
      attachments = attachments->next;
      return;
    }
   old = attachments;
   for (att=attachments->next; att; att=att->next) {
       if (att->object == remover) {
          old->next = att->next;
          return;
       }
       old = att;
   }
}


void
AnimObject::propagateAttachment(AnimObject *fromobj, Attachment att, 
        AnimObject *starter)
{
   Loc        *deptloc, *indeptloc;
   double      xdiff, ydiff, currx, curry;
   Attachment  mine;

   if (attachDoneFrom == view->GetUpdatingObject())  
      return;            // Already propagated through this AnimObject.  
                         // Short-circuit here.

   DamageIt();
   deptloc = Where(att->dept);
   indeptloc = fromobj->Where(att->indept);
   xdiff = indeptloc->XCoord() - deptloc->XCoord();
   ydiff = indeptloc->YCoord() - deptloc->YCoord();
   object->MyLoc(currx, curry);
   object->Warp(currx+xdiff+(att->xoffset), curry+ydiff+(att->yoffset));
   DamageIt();

   attachDoneFrom = starter;

   for (mine=attachments; mine; mine=mine->next)
       mine->object->propagateAttachment(this,mine,starter);

   attachDoneFrom = NULL;
}



// This takes an ActionInfoPtr out of an object's list of Actions, adjusts
// the list to make sure it's still correct (fixing first and last), deletes 
// the Action's memory, and then returns the next element in the list
//
struct ActionInfo *
AnimObject::DeleteActionInfoPtr(struct ActionInfo *temp)
{
   ActionInfoPtr old;

   if (temp == first)
      { first = temp->next;
        if (temp != last)
           first->prev = NULL;
        else
           last = NULL;
        old = temp;
        temp = first;
      }
   else
      { if (temp != last)
           temp->next->prev = temp->prev;
        else
           last = temp->prev;
        temp->prev->next = temp->next;
        old = temp;
        temp = temp->next;
      }
   if (old->action)
      delete old->action;  // removes offsets through Action's destructor
   delete old;
   return(temp);
}




/* This walks the list of Continuous Actions, checking for any with the ID
   of the given Action.  If it finds any, it removes them from the list,
   adjusts the list to make sure it's still correct (fixing first and last),
   deletes the ActionInfoPtr's memory, and then returns */

void
AnimObject::deleteContinActionInfoPtr(ActionPtr ap)
{
   ActionInfoPtr old,runner;
   
   runner = firstContinuous;
   while (runner) 
      if (runner->originalAction == ap) {  // got a match
         if (runner == firstContinuous)
            { firstContinuous = runner->next;
              if (runner != lastContinuous)
                 firstContinuous->prev = NULL;
              else
                 lastContinuous = NULL;
              old = runner;
              runner = firstContinuous;
              if (old->action)
                 delete old->action;
              delete old;
            }
         else  // runner != firstContinuous (in middle of list somewhere)
            { if (runner != lastContinuous)
                 runner->next->prev = runner->prev;
              else
                 lastContinuous = runner->prev;
              runner->prev->next = runner->next;
              old = runner;
              runner = runner->next;
              if (old->action)
                 delete old->action;
              delete old;
            }
         }
      else  // no match
         runner = runner->next;
}





int
AnimObject::doDeletes(int ti)
{
   if (deleters->value <= ti) {  // list is in sorted order
      object->DamageIt();
      view->RemoveUpdateObject(this->viewAOP);
      updating = 0;
      deleteMe();
      delete object;
      object = 0;   
      // deleter node will be freed by deleteMe()
      return(1);
   }
   return(0);
}



void
AnimObject::doContDeletes(int ti)
{
   ActionDeletePtr oldadp;

   while (continDeleters) {
      if (continDeleters->value <= ti) {  // list is in sorted order
         deleteContinActionInfoPtr(continDeleters->act);
         oldadp = continDeleters;
         continDeleters = continDeleters->next;
         delete oldadp;
         }
      else
         break;
   }
}



void
AnimObject::doChanges(int ti)
{
   ImplPtr c,old;
   double x,y;

   for (c=changers; c && (c->time <= ti) ; ) {  // list is in sorted order
      object->DamageIt();
      type = c->objimpl->Type();
      view = c->objimpl->MyView();
      if (c->existingpos) {   // use exisiting object's pos, not one
         object->MyLoc(x,y);   // given when new impl created.
         c->objimpl->Warp(x,y);
         }
      delete object;
      object = c->objimpl; 
      old = c;
      c = changers = c->next;
      delete old;
      object->DamageIt();
   }
}



int
AnimObject::Intercepts(double dlx, double dby, double drx, double dty)
{
   double lx,by,rx,ty;

   object->BoundBox(&lx,&by,&rx,&ty);
   if (lx > drx) return(0);
   if (rx < dlx) return(0);
   if (ty < dby) return(0);
   if (by > dty) return(0);
   return(1);
}



void 
AnimObject::IsInSet(AnimObject *set)
{
   AOPtr newone = new AONode;  
   newone->object = set;
   newone->next = objs;
   newone->prev = NULL;
   if (objs)
     objs->prev = newone;
   objs = newone;
}


void
AnimObject::SetIsGone(AnimObject *gone)
{
   AOPtr ao;    // a set this object was in is gone, so remove the ref

   for (ao=objs; ao; ao=ao->next) {
     if (ao->object == gone) {
        if (ao->next)
           ao->next->prev = ao->prev;
        if (ao->prev)
           ao->prev->next = ao->next;
        if (ao == objs)
           objs = ao->next;
        delete(ao);
        return;
      }
   }
}
  



Set::Set(View *v, int num, AnimObject *ao[]) : AnimObject( new SetImpl(v) )
{
   int i;
   AOPtr o;
   Set *set;

   type = P_Set;

   for (i=0; i<num; i++)
     { if (ao[i]->Type() != P_Set)
         another(ao[i]);
       else {
         set = (Set *) ao[i];
         for (o=set->Objs(); o; o=o->next)
           another(o->object);
       }
     }
}



Set::Set(const Set& s) : AnimObject( s )
{
   AOPtr a,newone,last1;

   type = P_Set;
   last1 = NULL;
   for (a = s.objs; a; a=a->next) {
      newone = new AONode;
      newone->object = a->object;
      newone->next = NULL;
      newone->prev = last1;
      if (objs)
        last1->next = newone;
      else
        objs = newone;
      last1 = a;    
      newone->object->IsInSet(this);  // put ref from obj to this set
   }
}


Set::~Set()
{
   AOPtr a,old;

   a = objs;
   while (a) {
       a->object->SetIsGone(this);  // removes ref from obj to this set
       old = a;
       a = a->next;
       delete(old);
   }
}
  
      
void
Set::RemoveFromSet(AnimObject *gone)
{
   AOPtr ao;    // an AnimObject was deleted, so get it out of this set

   for (ao=objs; ao; ao=ao->next) {
     if (ao->object == gone) {
        if (ao->next)
           ao->next->prev = ao->prev;
        if (ao->prev)
           ao->prev->next = ao->next;
        if (ao == objs)
           objs = ao->next;
        delete(ao);
        return;
      }
   }
}
  


void
Set::another(AnimObject *ao)
{
   AOPtr o,last1;

   o = objs;
   last1 = NULL;
   while (o) {
     if (o->object == ao)  // this object is already in the set, forget it
        return;
     last1 = o;
     o = o->next;
   }
   
   AOPtr newone = new AONode;  // not already in, so add it
   newone->object = ao;
   newone->next = NULL;
   newone->prev = last1;
   if (objs)
     last1->next = newone;
   else
     objs = newone;

   ao->IsInSet(this);  // designate that the object is in this set
}





AnimObjectImpl::AnimObjectImpl(View *vi, int v, double x, double y)
{
   if (vi->debug)
      printf("DEBUG: AnimObjectImpl::AnimObjectImpl vis=%d  x=%f y=%f\n",v,x,y);

   view = vi;
   type = P_Misc;
   visibility = v;
   locx = x;
   locy = y;
}



AnimObjectImpl::AnimObjectImpl(const AnimObjectImpl& a)
{
   view =  a.view;
   type = a.type;
   visibility = a.visibility;
   locx = a.locx;
   locy = a.locy;
}





void
AnimObjectImpl::Trans(char* type1, double dx, double dy)
{
   // This should never get called with a Set AnimObject

   AOPtr aop;

   if (view->debug)
      printf("DEBUG: AnimObjectImpl::Trans  %s (%f, %f)\n",type,dx,dy);

   if (streql(type1,"MOVE"))
      { 
        DamageIt();
        locx += dx;
        locy += dy;
        DamageIt();
      }
   else if (streql(type1,"VIS"))
      { visibility = !visibility;
        DamageIt();
      }
   else if (streql(type1,"RAISE"))
      { for (aop=view->AOHead; aop->object != wrapper; aop=aop->next) ;
        if (!aop) // weird error
           return;
        if (aop == view->AOHead) // do nothing
           return;
        else if (aop == view->AOTail)
           view->AOTail = aop->prev;
        else
           aop->next->prev = aop->prev;
        aop->prev->next = aop->next;
        aop->next = view->AOHead;
        view->AOHead->prev = aop;
        view->AOHead = aop;
        aop->prev = NULL;

        DamageIt();
      }
   else if (streql(type1,"LOWER"))
      { for (aop=view->AOHead; aop->object != wrapper; aop=aop->next) ;
        if (!aop) // weird error
           return;
        if (aop == view->AOTail) // do nothing
           return;
        else if (aop == view->AOHead)
           view->AOHead = aop->next;
        else
           aop->prev->next = aop->next;
        aop->next->prev = aop->prev;
        aop->prev = view->AOTail;
        view->AOTail->next = aop;
        view->AOTail = aop;
        aop->next = NULL;

        DamageIt();
      }
   else if (streql(type1,"ALTER_LL"))
      { 
        view->AlterLLCoord(dx,dy);
        view->RefreshNeeded(1);
      }
   else if (streql(type1,"ALTER_UR"))
      { 
        view->AlterURCoord(dx,dy);
        view->RefreshNeeded(1);
      }
   else
      { transSpecial(type1,dx,dy);
      }
}



void
AnimObjectImpl::DamageIt()
{
   double lx,by,rx,ty;

   BoundBox(&lx,&by,&rx,&ty);         // where this object is now
// printf("Damaging  (%.3lf, %.3lf)   (%.3lf, %.3lf)\n",lx,by,rx,ty);
   view->DamageCheck(lx,by,rx,ty);    // update view's damage region
}



Line& 
Line::operator=(const Line& rhs) 
{
   if  (this == &rhs) return *this;
   AnimObject::operator=(rhs);
   return *this;
}




LineImpl::LineImpl(View *an, int v, double lx, double ly, 
           double sx, double sy, COLOR c, double w, double s, int a) 
          :  AnimObjectImpl(an,v,lx,ly) 
{
   type = P_Line;
   sizex = sx;
   sizey = sy;
   strcpy(colorname,c);
   color = load_color(colorname);
   width = w;
   style = s;
   arrow = a;
   /* figure out arrow here */
   if ((arrow == 1) || (arrow == 3))
      compute_arrow_off(sx,sy,&arrowf1x,&arrowf1y,&arrowf2x,&arrowf2y);
   if ((arrow == 2) || (arrow == 3))
      compute_arrow_off(-sx,-sy,&arrowf3x,&arrowf3y,&arrowf4x,&arrowf4y);
}


LineImpl::LineImpl(const LineImpl& l) : AnimObjectImpl( l )
{
   sizex = l.sizex;
   sizey = l.sizey;
   strcpy(colorname,l.colorname);
   color = l.color;  // know it's already been loaded
   width = l.width;
   style = l.style;
   arrow = l.arrow;
   /* figure out arrow here */
   if ((arrow == 1) || (arrow == 3)) {
      arrowf1x = l.arrowf1x;
      arrowf1y = l.arrowf1y;
      arrowf2x = l.arrowf2x;
      arrowf2y = l.arrowf2y;
      }
   if ((arrow == 2) || (arrow == 3)) {
      arrowf3x = l.arrowf3x;
      arrowf3y = l.arrowf3y;
      arrowf4x = l.arrowf4x;
      arrowf4y = l.arrowf4y;
      }
}
   


LineImpl::~LineImpl()
{
   Erase();
}


void
LineImpl::GetValues(View* *an, int* v, double* lx, double* ly, 
           double* sx, double* sy, COLOR c, double* w, double* s, int* a) 
{
   if (an) *an=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (sx) *sx=sizex;
   if (sy) *sy=sizey;
   if (c) strcpy(c,colorname);
   if (w) *w=width;
   if (s) *s=style;
   if (a) *a=arrow;
}


LocPtr
LineImpl::Where(PART p)
{
   double	   xmin,xmax,ymin,ymax;

   BoundBox(&xmin,&ymin,&xmax,&ymax);

   switch (p)
   {
      case PART_NW :
	 return( new Loc(xmin,ymax) );
      case PART_N :
	 return( new Loc((xmin+xmax)/2.0,ymax) );
      case PART_NE :
	 return( new Loc(xmax,ymax) );
      case PART_E :
	 return( new Loc(xmax,(ymin+ymax)/2.0) );
      case PART_SE :
	 return( new Loc(xmax,ymin) );
      case PART_S :
	 return( new Loc((xmin+xmax)/2.0,ymin) );
      case PART_SW :
	 return( new Loc(xmin,ymin) );
      case PART_W :
	 return( new Loc(xmin,(ymin+ymax)/2.0) );
      case PART_C :
	 return( new Loc((xmin+xmax)/2.0,(ymin+ymax)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}



void
LineImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   double ax0,ax1,ay0,ay1;

   *lx = (sizex > ZERO ? locx : locx+sizex);
   *rx = (sizex > ZERO ? locx+sizex : locx);
   *ty = (sizey > ZERO ? locy+sizey : locy);
   *by = (sizey > ZERO ? locy: locy+sizey);
   if ((arrow & 0x1) != 0)    // below is dumb, but it works
      {        // need to take arrows into account on boundbox
        ax0 = locx+sizex+arrowf1x;
        ax1 = locx+sizex+arrowf2x;
        if (ax0 < *lx) *lx = ax0;
        if (ax1 < *lx) *lx = ax1;
        if (ax0 > *rx) *rx = ax0;
        if (ax1 > *rx) *rx = ax1;
        ay0 = locy+sizey+arrowf1y;
        ay1 = locy+sizey+arrowf2y;
        if (ay0 < *by) *by = ay0;
        if (ay1 < *by) *by = ay1;
        if (ay0 > *ty) *ty = ay0;
        if (ay1 > *ty) *ty = ay1;
      }
   if ((arrow & 0x2) != 0)
      { 
        ax0 = locx+arrowf3x;
        ax1 = locx+arrowf4x;
        if (ax0 < *lx) *lx = ax0;
        if (ax1 < *lx) *lx = ax1;
        if (ax0 > *rx) *rx = ax0;
        if (ax1 > *rx) *rx = ax1;
        ay0 = locy+arrowf3y;
        ay1 = locy+arrowf4y;
        if (ay0 < *by) *by = ay0;
        if (ay1 < *by) *by = ay1;
        if (ay0 > *ty) *ty = ay0;
        if (ay1 > *ty) *ty = ay1;
      }
}






void
LineImpl::transSpecial(char *type1, double dx, double dy)
{
   if (streql(type1,"FILL"))
      { width +=  dx;
	if (width < 0.0)
	   width = 0.0;
	else if (width > 1.0)
	   width = 1.0;
        style +=  dy;
	if (style < 0.0)
	   style = 0.0;
	else if (style > 1.0)
	   style = 1.0;
        DamageIt();
      }
   else if (streql(type1,"RESIZE"))
      { DamageIt();
        sizex += dx;
	sizey += dy;
        if ((arrow == 1) || (arrow == 3))
           compute_arrow_off(sizex,sizey,&arrowf1x,&arrowf1y,
                                &arrowf2x,&arrowf2y);
        if ((arrow == 2) || (arrow == 3))
           compute_arrow_off(-sizex,-sizey,&arrowf3x,&arrowf3y,
                                &arrowf4x,&arrowf4y);
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
}

   

void
LineImpl::Draw()
{
   drawer(color);
}

void
LineImpl::Erase()
{
   drawer(view->GetBgColor());
}



void
LineImpl::drawer(COLORINDEX col)
{
   int x0,y0,x1,y1,ax0,ay0,ax1,ay1;
   LINE_STYLE lstyle,old = 0;

   if (!visibility)
      return;

   x0 = view->TO_PIX_X(locx);
   y0 = view->TO_PIX_Y(locy);
   x1 = view->TO_PIX_X(locx+sizex);
   y1 = view->TO_PIX_Y(locy+sizey);

   set_color(col);
   lstyle = get_linestyle(width,style);
   if (lstyle != SOLID)
      old = line_style(lstyle);
   XDrawLine(_display,view->DrawWindow(),inq_gc(),x0,y0,x1,y1);

   if ((arrow & 0x1) != 0)
      { 
        ax0 = view->TO_PIX_X(locx+sizex+arrowf1x);
        ay0 = view->TO_PIX_Y(locy+sizey+arrowf1y);
        ax1 = view->TO_PIX_X(locx+sizex+arrowf2x);
        ay1 = view->TO_PIX_Y(locy+sizey+arrowf2y);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),x1,y1,ax0,ay0);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),x1,y1,ax1,ay1);
      }
   if ((arrow & 0x2) != 0)
      { 
        ax0 = view->TO_PIX_X(locx+arrowf3x);
        ay0 = view->TO_PIX_Y(locy+arrowf3y);
        ax1 = view->TO_PIX_X(locx+arrowf4x);
        ay1 = view->TO_PIX_Y(locy+arrowf4y);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),x0,y0,ax0,ay0);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),x0,y0,ax1,ay1);
      }

   if (lstyle != SOLID)
      line_style(old);
}





Rectangle& 
Rectangle::operator=(const Rectangle& rhs) 
{
   if  (this == &rhs) return *this;
   AnimObject::operator=(rhs);
   return *this;
}





RectangleImpl::RectangleImpl(View *a, int v, double lx, double ly, 
              double sx, double sy, COLOR c, double f) 
              : AnimObjectImpl(a,v,lx,ly)
{
   type = P_Rectangle;
   sizex = sx;
   if (sx < 0.0) {
      cerr << "Warning:  Rectangle with x_size less than 0.0" << endl;
      sizex = 0.0;
      }
   sizey = sy;
   if (sy < 0.0) {
      cerr << "Warning:  Rectangle with y_size less than 0.0" << endl;
      sizey = 0.0;
      }
   strcpy(colorname,c);
   color = load_color(colorname);
   fill = f;
}


RectangleImpl::RectangleImpl(const RectangleImpl& r) : AnimObjectImpl( r )
{
   sizex = r.sizex;
   sizey = r.sizey;
   strcpy(colorname,r.colorname);
   color = r.color;
   fill = r.fill;
}
  


RectangleImpl::~RectangleImpl()
{
   Erase();
}



void 
RectangleImpl::GetValues(View* *vi, int *v, double *lx, double *ly, 
          double *sx, double *sy, COLOR c, double *f)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (sx) *sx=sizex;
   if (sy) *sy=sizey;
   if (c) strcpy(c,colorname);
   if (f) *f=fill;
}




LocPtr
RectangleImpl::Where(PART p)
{
   double	   x0,y0,x1,y1;

   BoundBox(&x0,&y1,&x1,&y0);

   switch (p)
   {
      case PART_C :
	 return( new Loc((x0 + x1)/2.0,(y0 + y1)/2.0) );
      case PART_NW :
	 return( new Loc(x0,y0) );
      case PART_N :
	 return( new Loc((x0+ x1)/2.0,y0) );
      case PART_NE :
	 return( new Loc(x1,y0) );
      case PART_E :
	 return( new Loc(x1,(y0 + y1)/2.0) );
      case PART_SE :
	 return( new Loc(x1,y1) );
      case PART_S :
	 return( new Loc((x0 + x1)/2.0,y1) );
      case PART_SW :
	 return( new Loc(x0,y1) );
      case PART_W :
	 return( new Loc(x0,(y0 + y1)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}



void
RectangleImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   *lx= locx;
   *rx= *lx + sizex;
   *by= locy;
   *ty= *by + sizey;
}




void
RectangleImpl::transSpecial(char *type1, double dx, double dy)
{
   if (streql(type1,"FILL"))
      { fill +=  dx;
	if (fill < 0.0)
	   fill = 0.0;
	else if (fill > 1.0)
	   fill = 1.0;
        DamageIt();
      }
   else if (streql(type1,"RESIZE"))
      { DamageIt();
        sizex += dx;
	sizey += dy;
	if (sizex < ZERO)
	   sizex = ZERO;
	if (sizey < ZERO)
	   sizey = ZERO;
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
}

   


void
RectangleImpl::Draw()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   drawer(color,curfill);
}

void
RectangleImpl::Erase()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   if (curfill == POLKA_FILL_OUTLINE)
      drawer(view->GetBgColor(),POLKA_FILL_OUTLINE);
   else
      drawer(view->GetBgColor(),POLKA_FILL_SOLID);
}


void
RectangleImpl::drawer(COLORINDEX col, FILL_STYLE curfill)
{
   int x0,y0,xs,ys;
   FILL_STYLE oldfill;

   if (!visibility)
      return;

   x0 = view->TO_PIX_X(locx);
   y0 = view->TO_PIX_Y(locy);
   xs = view->SIZE_PIX_X(sizex);
   ys = view->SIZE_PIX_Y(sizey);

   set_color(col);
   if (curfill == POLKA_FILL_OUTLINE)
      {
        XDrawRectangle(_display,view->DrawWindow(),inq_gc(),x0,y0-ys,xs,ys);
      }
   else
      {
        oldfill = fill_style(curfill);
        XFillRectangle(_display,view->DrawWindow(),inq_gc(),x0,y0-ys,xs,ys);
        fill_style(oldfill);
      }
}



Circle& 
Circle::operator=(const Circle& rhs) 
{
  if  (this == &rhs) return *this;
  AnimObject::operator=(rhs);
  return *this;
}




CircleImpl::CircleImpl(View *a, int v, double lx, double ly, double r,
                  COLOR c, double f) 
          : AnimObjectImpl(a,v,lx,ly)
{
   type = P_Circle;
   radius = r;
   strcpy(colorname,c);
   color = load_color(colorname);
   fill = f;
}


CircleImpl::CircleImpl(const CircleImpl& c) : AnimObjectImpl( c )
{
   radius = c.radius;
   strcpy(colorname,c.colorname);
   color = c.color;
   fill = c.fill;
}
  

CircleImpl::~CircleImpl()
{
   Erase();
}


void 
CircleImpl::GetValues(View* *vi, int *v, double *lx, double *ly, double *r, 
              COLOR c, double *f)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (r) *r=radius;
   if (c) strcpy(c,colorname);
   if (f) *f=fill;
}   



LocPtr
CircleImpl::Where(PART p)
{
   double	   x0,y0,corner;

   x0 = locx;
   y0 = locy;
   corner = sqrt( (radius * radius) / 2.0 );
   switch (p)
   {
      case PART_C :
	 return( new Loc(x0,y0) );
      case PART_NW :
	 return( new Loc(x0-corner,y0+corner) );
      case PART_N :
	 return( new Loc(x0,y0+radius) );
      case PART_NE :
	 return( new Loc(x0+corner,y0+corner) );
      case PART_E :
	 return( new Loc(x0+radius,y0) );
      case PART_SE :
	 return( new Loc(x0+corner,y0-corner) );
      case PART_S :
	 return( new Loc(x0,y0-radius) );
      case PART_SW :
	 return( new Loc(x0-corner,y0-corner) );
      case PART_W :
	 return( new Loc(x0-radius,y0) );
   }
   return( new Loc(0.0,0.0) );
}



void
CircleImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   *lx = locx - radius;
   *rx = locx + radius;
   *ty = locy + radius;
   *by = locy - radius;
}




void
CircleImpl::transSpecial(char *type1, double dx, double dy)
{
   if (streql(type1,"FILL"))
      { fill +=  dx;
	if (fill < 0.0)
	   fill = 0.0;
	else if (fill > 1.0)
	   fill = 1.0;
        DamageIt();
      }
   else if (streql(type1,"RESIZE"))
      { DamageIt();
        radius += dx;
	if (radius < ZERO)
	   radius = ZERO;
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
   }
}

   


void
CircleImpl::Draw()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   drawer(color,curfill);
}

void
CircleImpl::Erase()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   if (curfill == POLKA_FILL_OUTLINE)
      drawer(view->GetBgColor(),POLKA_FILL_OUTLINE);
   else
      drawer(view->GetBgColor(),POLKA_FILL_SOLID);
}


void
CircleImpl::drawer(COLORINDEX col,FILL_STYLE curfill)
{
   int x0,y0,radx,rady;
   FILL_STYLE oldfill;

   if (!visibility)
      return;

   x0 = view->TO_PIX_X(locx);
   y0 = view->TO_PIX_Y(locy);
   radx = view->SIZE_PIX_X(radius);
   rady = view->SIZE_PIX_Y(radius);
   set_color(col);

   if (curfill == POLKA_FILL_OUTLINE)
      {
        XDrawArc(_display,view->DrawWindow(),inq_gc(),x0-radx,y0-rady,
                                  radx<<1,rady<<1,0,23040); 
      }                                /* multiply by 2 to get size */
   else 
      {
        oldfill = fill_style(curfill);
        XFillArc(_display,view->DrawWindow(),inq_gc(),x0-radx,y0-rady,
                      radx<<1,rady<<1,0,23040);  // 23040 == 64*360
        fill_style(oldfill);
      }
}




Ellipse& 
Ellipse::operator=(const Ellipse& rhs) 
{
  if  (this == &rhs) return *this;
  AnimObject::operator=(rhs);
  return *this;
}




EllipseImpl::EllipseImpl(View *a, int v, double lx, double ly, double sx, 
                   double sy, COLOR c, double f) 
             : AnimObjectImpl(a,v,lx,ly)
{
   type = P_Ellipse;
   sizex = sx;
   sizey = sy;
   strcpy(colorname,c);
   color = load_color(colorname);
   fill = f;
}

EllipseImpl::EllipseImpl(const EllipseImpl& e) : AnimObjectImpl( e )
{
   sizex = e.sizex;
   sizey = e.sizey;
   strcpy(colorname,e.colorname);
   color = e.color;
   fill = e.fill;
}



EllipseImpl::~EllipseImpl()
{
   Erase();
}


void 
EllipseImpl::GetValues(View* *vi, int *v, double *lx, double *ly, 
                double *sx, double *sy, COLOR c, double *f)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (sx) *sx=sizex;
   if (sy) *sy=sizey;
   if (c) strcpy(c,colorname);
   if (f) *f=fill;
}




LocPtr
EllipseImpl::Where(PART p)
{
   double	   cx,cy,rx,ry;

   cx = locx;
   cy = locy;
   rx = sizex;
   ry = sizey;
   switch (p)
   {
      case PART_C :
	 return( new Loc(cx,cy) );
      case PART_NW :
	 return( new Loc(cx-rx,cy+ry) );
      case PART_N :
	 return( new Loc(cx,cy+ry) );
      case PART_NE :
	 return( new Loc(cx+rx,cy+ry) );
      case PART_E :
	 return( new Loc(cx+rx,cy) );
      case PART_SE :
	 return( new Loc(cx+rx,cy-ry) );
      case PART_S :
	 return( new Loc(cx,cy-ry) );
      case PART_SW :
	 return( new Loc(cx-rx,cy-ry) );
      case PART_W :
	 return( new Loc(cx-rx,cy) );
   }
   return( new Loc(0.0,0.0) );
}



void
EllipseImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   *lx = locx - sizex;
   *rx = locx + sizex;
   *ty = locy + sizey;
   *by = locy - sizey;
}




void
EllipseImpl::transSpecial(char *type1, double dx, double dy)
{
   if (streql(type1,"FILL"))
      { fill +=  dx;
	if (fill < 0.0)
	   fill = 0.0;
	else if (fill > 1.0)
	   fill = 1.0;
        DamageIt();
      }
   else if (streql(type1,"RESIZE"))
      { DamageIt();
        sizex += dx;
	sizey += dy;
	if (sizex < ZERO)
	   sizex = ZERO;
	if (sizey < ZERO)
	   sizey = ZERO;
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
}

   


void
EllipseImpl::Draw()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   drawer(color,curfill);
}

void
EllipseImpl::Erase()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   if (curfill == POLKA_FILL_OUTLINE)
      drawer(view->GetBgColor(),POLKA_FILL_OUTLINE);
   else
      drawer(view->GetBgColor(),POLKA_FILL_SOLID);
}


void
EllipseImpl::drawer(COLORINDEX col, FILL_STYLE curfill)
{
   int x0,y0,x1,y1;
   FILL_STYLE oldfill;

   if (!visibility)
      return;

   x0 = view->TO_PIX_X(locx);
   y0 = view->TO_PIX_Y(locy);
   x1 = view->SIZE_PIX_X(sizex);
   y1 = view->SIZE_PIX_Y(sizey);

   set_color(col);

   if (curfill == POLKA_FILL_OUTLINE)
      {
        XDrawArc(_display,view->DrawWindow(),inq_gc(),x0-x1,y0-y1,
                        x1<<1,y1<<1,0,23040);  // 23040 == 64*360
      }                                /* multiply by 2 to get size */
   else 
      {
        oldfill = fill_style(curfill);
        XFillArc(_display,view->DrawWindow(),inq_gc(),x0-x1,y0-y1,
                               x1<<1,y1<<1,0,23040);
        fill_style(oldfill);
      }
}




Polyline& 
Polyline::operator=(const Polyline& rhs) 
{
  if  (this == &rhs) return *this;
  AnimObject::operator=(rhs);
  return *this;
}




PolylineImpl::PolylineImpl(View *an, int v, double lx, double ly, int ve, 
               double vtx[], double vty[], COLOR c, double w, double s, 
               int a) 
               : AnimObjectImpl(an,v,lx,ly)
{
   int i;
   double fx,fy,bx,by;

   type = P_Polyline;
   vertices = ve;
   for (i=0; i<vertices-1; i++) {
       vx[i] = vtx[i];
       vy[i] = vty[i];
     }
   strcpy(colorname,c);
   color = load_color(colorname);
   width = w;
   style = s;
   arrow = a;
   if (arrow != 0)
      arrow_poly_direction(vertices,vx,vy,&fx,&fy,&bx,&by);
   if ((arrow == 1) || (arrow == 3))
      compute_arrow_off(fx,fy,&arrowf1x,&arrowf1y,&arrowf2x,&arrowf2y);
   if ((arrow == 2) || (arrow == 3))
      compute_arrow_off(bx,by,&arrowf3x,&arrowf3y,&arrowf4x,&arrowf4y);
   /* do arrows here */
}


PolylineImpl::PolylineImpl(const PolylineImpl& p) : AnimObjectImpl( p )
{
   vertices = p.vertices;
   for (int i=0; i<vertices-1; i++) {
       vx[i] = p.vx[i];
       vy[i] = p.vy[i];
     }
   strcpy(colorname,p.colorname);
   color = p.color;
   width = p.width;
   style = p.style;
   arrow = p.arrow;
   if ((arrow == 1) || (arrow == 3)) {
      arrowf1x = p.arrowf1x;
      arrowf1y = p.arrowf1y;
      arrowf2x = p.arrowf2x;
      arrowf2y = p.arrowf2y;
      }
   if ((arrow == 2) || (arrow == 3)) {
      arrowf3x = p.arrowf3x;
      arrowf3y = p.arrowf3y;
      arrowf4x = p.arrowf4x;
      arrowf4y = p.arrowf4y;
      }
}



PolylineImpl::~PolylineImpl()
{
   Erase();
}   



void 
PolylineImpl::GetValues(View* *vi, int *v, double *lx, double *ly, int *ve, 
                double vtx[], double vty[], COLOR c, 
                double *w, double *s, int *a)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (ve) *ve=vertices;
   for (int i=0; i<vertices; i++) {
       if (vtx) vtx[i] = vx[i];
       if (vty) vty[i] = vy[i];
   }
   if (c) strcpy(c,colorname);
   if (w) *w=width;
   if (s) *s=style;
   if (a) *a=arrow;
}


LocPtr
PolylineImpl::Where(PART p)
{
   double	   xmin,ymin,xmax,ymax;

   BoundBox(&xmin,&ymin,&xmax,&ymax);

   switch (p)
   {
      case PART_C :
	 return( new Loc((xmin + xmax)/2.0,(ymin + ymax)/2.0) );
      case PART_NW :
	 return( new Loc(xmin,ymax) );
      case PART_N :
	 return( new Loc((xmin+ xmax)/2.0,ymax) );
      case PART_NE :
	 return( new Loc(xmax,ymax) );
      case PART_E :
	 return( new Loc(xmax,(ymin + ymax)/2.0) );
      case PART_SE :
	 return( new Loc(xmax,ymin) );
      case PART_S :
	 return( new Loc((xmin + xmax)/2.0,ymin) );
      case PART_SW :
	 return( new Loc(xmin,ymin) );
      case PART_W :
	 return( new Loc(xmin,(ymin + ymax)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}



void
PolylineImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   double x,y,xmin,xmax,ymin,ymax,tox,toy,ax0,ax1,ay0,ay1;
   int i;

   xmin = xmax = locx;
   ymin = ymax = locy;   /* Find the bounding box */
   for (i=0; i<(vertices - 1); ++i)
      { if ((x = (locx+vx[i])) > xmax) xmax = x;
	if ((x = (locx+vx[i])) < xmin) xmin = x;
	if ((y = (locy+vy[i])) > ymax) ymax = y;
	if ((y = (locy+vy[i])) < ymin) ymin = y;
      }
   *lx = xmin; *rx = xmax; *ty = ymax; *by = ymin;
   if ((arrow & 0x1) != 0)    // below is dumb, but it works
      {        // need to take arrows into account on boundbox
        tox = locx + vx[vertices-2];
        toy = locy + vy[vertices-2];
        ax0 = tox+arrowf1x;
        ax1 = tox+arrowf2x;
        if (ax0 < *lx) *lx = ax0;
        if (ax1 < *lx) *lx = ax1;
        if (ax0 > *rx) *rx = ax0;
        if (ax1 > *rx) *rx = ax1;
        ay0 = toy+arrowf1y;
        ay1 = toy+arrowf2y;
        if (ay0 < *by) *by = ay0;
        if (ay1 < *by) *by = ay1;
        if (ay0 > *ty) *ty = ay0;
        if (ay1 > *ty) *ty = ay1;
      }
   if ((arrow & 0x2) != 0)
      { 
        ax0 = locx+arrowf3x;
        ax1 = locx+arrowf4x;
        if (ax0 < *lx) *lx = ax0;
        if (ax1 < *lx) *lx = ax1;
        if (ax0 > *rx) *rx = ax0;
        if (ax1 > *rx) *rx = ax1;
        ay0 = locy+arrowf3y;
        ay1 = locy+arrowf4y;
        if (ay0 < *by) *by = ay0;
        if (ay1 < *by) *by = ay1;
        if (ay0 > *ty) *ty = ay0;
        if (ay1 > *ty) *ty = ay1;
      }
}




void
PolylineImpl::transSpecial(char *type1, double dx, double dy)
{
   int i,v,recompute=0;
   double fx,fy,bx,by;

   if (streql(type1,"FILL"))
      { width +=  dx;
	if (width < 0.0)
	   width = 0.0;
	else if (width > 1.0)
	   width = 1.0;
        style +=  dy;
	if (style < 0.0)
	   style = 0.0;
	else if (style > 1.0)
	   style = 1.0;
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
   else if (streqln(type1,"RESIZE",6)) {
      if (('1' <= *(type1+6)) && (*(type1+6) <= '7')) {
         DamageIt();
         recompute = 1;
         v = (int) (*(type1+6) - '0');
         for (i=v-1; i<vertices-1; i++) {
            vx[i] += dx;
            vy[i] += dy;
            }
         DamageIt();
         }
      }      
   else if (streqln(type1,"GRAB",4)) {
      if (('1' <= *(type1+4)) && (*(type1+4) <= '7')) {
         DamageIt();
         recompute = 1;
         v = (int) (*(type1+4) - '1');
         vx[v] += dx;
         vy[v] += dy;
         DamageIt();
         }
      }      
   if (recompute) {
      if (arrow != 0)
         arrow_poly_direction(vertices,vx,vy,&fx,&fy,&bx,&by);
      if ((arrow == 1) || (arrow == 3))
         compute_arrow_off(fx,fy,&arrowf1x,&arrowf1y,&arrowf2x,&arrowf2y);
      if ((arrow == 2) || (arrow == 3))
         compute_arrow_off(bx,by,&arrowf3x,&arrowf3y,&arrowf4x,&arrowf4y);
      }
}

   


void
PolylineImpl::Draw()
{
   drawer(color);
}

void
PolylineImpl::Erase()
{
   drawer(view->GetBgColor());
}


void
PolylineImpl::drawer(COLORINDEX col)
{
   int i,ax0,ay0,ax1,ay1;
   double tx,ty;
   LINE_STYLE lstyle,old = 0;
   XPoint pts[10];

   if (!visibility)
      return;

   pts[0].x = view->TO_PIX_X(locx);
   pts[0].y = view->TO_PIX_Y(locy);
   for (i=1; i<vertices; i++) {
      pts[i].x = view->TO_PIX_X(locx + vx[i-1]);
      pts[i].y = view->TO_PIX_Y(locy + vy[i-1]);
      }
   set_color(col);
   lstyle = get_linestyle(width,style);
   if (lstyle != SOLID)
      old = line_style(lstyle);
   XDrawLines(_display,view->DrawWindow(),inq_gc(),pts,vertices,CoordModeOrigin);

   if ((arrow & 0x1) != 0)
      { tx = locx + vx[vertices-2];
        ty = locy + vy[vertices-2];
        ax0 = view->TO_PIX_X(tx+arrowf1x);
        ay0 = view->TO_PIX_Y(ty+arrowf1y);
        ax1 = view->TO_PIX_X(tx+arrowf2x);
        ay1 = view->TO_PIX_Y(ty+arrowf2y);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),
                            pts[vertices-1].x,pts[vertices-1].y,ax0,ay0);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),
                            pts[vertices-1].x,pts[vertices-1].y,ax1,ay1);
      }
   if ((arrow & 0x2) != 0)
      { 
        ax0 = view->TO_PIX_X(locx+arrowf3x);
        ay0 = view->TO_PIX_Y(locy+arrowf3y);
        ax1 = view->TO_PIX_X(locx+arrowf4x);
        ay1 = view->TO_PIX_Y(locy+arrowf4y);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),pts[0].x,pts[0].y,ax0,ay0);
        XDrawLine(_display,view->DrawWindow(),inq_gc(),pts[0].x,pts[0].y,ax1,ay1);
      }

   if (lstyle != SOLID)
      line_style(old);
}



Pie& 
Pie::operator=(const Pie& rhs) 
{
  if  (this == &rhs) return *this;
  AnimObject::operator=(rhs);
  return *this;
}




PieImpl::PieImpl(View *a, int v, double lx, double ly, double r,
                  double ba, double de, COLOR c, double f) 
          : AnimObjectImpl(a,v,lx,ly)
{
   type = P_Pie;
   radius = r;
   begangle = ba;
   delta = de;
   strcpy(colorname,c);
   color = load_color(colorname);
   fill = f;
}


PieImpl::PieImpl(const PieImpl& c) : AnimObjectImpl( c )
{
   radius = c.radius;
   begangle = c.begangle;
   delta = c.delta;
   strcpy(colorname,c.colorname);
   color = c.color;
   fill = c.fill;
}
  

PieImpl::~PieImpl()
{
   Erase();
}


void 
PieImpl::GetValues(View* *vi, int *v, double *lx, double *ly, 
                double *rad, double *bang, double *del, 
                COLOR c, double *f)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (rad) *rad=radius;
   if (bang) *bang=begangle;
   if (del) *del=delta;
   if (c) strcpy(c,colorname);
   if (f) *f=fill;
}


LocPtr
PieImpl::Where(PART p)
{
   double	   x0,y0,corner;

   x0 = locx;
   y0 = locy;
   corner = sqrt( (radius * radius) / 2.0 );
   switch (p)
   {
      case PART_C :
	 return( new Loc(x0,y0) );
      case PART_NW :
	 return( new Loc(x0-corner,y0+corner) );
      case PART_N :
	 return( new Loc(x0,y0+radius) );
      case PART_NE :
	 return( new Loc(x0+corner,y0+corner) );
      case PART_E :
	 return( new Loc(x0+radius,y0) );
      case PART_SE :
	 return( new Loc(x0+corner,y0-corner) );
      case PART_S :
	 return( new Loc(x0,y0-radius) );
      case PART_SW :
	 return( new Loc(x0-corner,y0-corner) );
      case PART_W :
	 return( new Loc(x0-radius,y0) );
   }
   return( new Loc(0.0,0.0) );
}



void
PieImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   *lx = locx - radius;
   *rx = locx + radius;
   *ty = locy + radius;
   *by = locy - radius;
}




void
PieImpl::transSpecial(char *type1, double dx, double dy)
{
   if (streql(type1,"FILL"))
      { fill +=  dx;
	if (fill < 0.0)
	   fill = 0.0;
	else if (fill > 1.0)
	   fill = 1.0;
        DamageIt();
      }
   else if (streql(type1,"GRAB"))
      { begangle += dx;
        delta += dy;
        DamageIt();
      }
   else if (streql(type1,"RESIZE"))
      { DamageIt();
        radius += dx;
	if (radius < ZERO)
	   radius = ZERO;
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
}

   


void
PieImpl::Draw()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   drawer(color,curfill);
}

void
PieImpl::Erase()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   if (curfill == POLKA_FILL_OUTLINE)
      drawer(view->GetBgColor(),POLKA_FILL_OUTLINE);
   else
      drawer(view->GetBgColor(),POLKA_FILL_SOLID);
}


void
PieImpl::drawer(COLORINDEX col, FILL_STYLE curfill)
{
   int x0,y0,radx,rady;
   FILL_STYLE oldfill;

   if (!visibility)
      return;

   x0 = view->TO_PIX_X(locx);
   y0 = view->TO_PIX_Y(locy);
   radx = view->SIZE_PIX_X(radius);
   rady = view->SIZE_PIX_Y(radius);
   set_color(col);

   XSetArcMode(_display, inq_gc(), ArcPieSlice);
   if (curfill == POLKA_FILL_OUTLINE)
      {
        // this should be fixed!
        XDrawArc(_display,view->DrawWindow(),inq_gc(),x0-radx,y0-rady,
                                  radx<<1,rady<<1,0,23040);
      }                                /* multiply by 2 to get size */
   else 
      {
        oldfill = fill_style(curfill);
        XFillArc(_display,view->DrawWindow(),inq_gc(),x0-radx,y0-rady,
                 radx<<1,rady<<1, int(begangle*23040), int(delta*23040));
        fill_style(oldfill);       // 23040 == 360*64
      }
}




Spline& 
Spline::operator=(const Spline& rhs) 
{
   if  (this == &rhs) return *this;
   AnimObject::operator=(rhs);
   return *this;
}



SplineImpl::SplineImpl(View *a, int v, double lx, double ly, int ve, 
              double vtx[], double vty[], COLOR c, double w, double s) 
               : AnimObjectImpl(a,v,lx,ly)
{
   int i;

   type = P_Spline;
   vertices = ve;
   for (i=0; i<vertices-1; i++) {
       vx[i] = vtx[i];
       vy[i] = vty[i];
     }
   strcpy(colorname,c);
   color = load_color(colorname);
   width = w;
   style = s;
}

SplineImpl::SplineImpl(const SplineImpl& s) : AnimObjectImpl( s )
{
   vertices = s.vertices;
   for (int i=0; i<vertices-1; i++) {
       vx[i] = s.vx[i];
       vy[i] = s.vy[i];
     }
   strcpy(colorname,s.colorname);
   color = s.color;
   width = s.width;
   style = s.style;
}




SplineImpl::~SplineImpl()
{
   Erase();
}


void 
SplineImpl::GetValues(View* *vi, int *v, double *lx, double *ly, int *ve, 
                double vtx[], double vty[], COLOR c, double *w, double *s)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (ve) *ve=vertices;
   for (int i=0; i<vertices; i++) {
       if (vtx) vtx[i] = vx[i];
       if (vty) vty[i] = vy[i];
   }
   if (c) strcpy(c,colorname);
   if (w) *w=width;
   if (s) *s=style;
}	



LocPtr
SplineImpl::Where(PART p)
{
   double	   xmin,ymin,xmax,ymax;

   BoundBox(&xmin,&ymin,&xmax,&ymax);

   switch (p)
   {
      case PART_C :
	 return( new Loc((xmin + xmax)/2.0,(ymin + ymax)/2.0) );
      case PART_NW :
	 return( new Loc(xmin,ymax) );
      case PART_N :
	 return( new Loc((xmin+ xmax)/2.0,ymax) );
      case PART_NE :
	 return( new Loc(xmax,ymax) );
      case PART_E :
	 return( new Loc(xmax,(ymin + ymax)/2.0) );
      case PART_SE :
	 return( new Loc(xmax,ymin) );
      case PART_S :
	 return( new Loc((xmin + xmax)/2.0,ymin) );
      case PART_SW :
	 return( new Loc(xmin,ymin) );
      case PART_W :
	 return( new Loc(xmin,(ymin + ymax)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}



void
SplineImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   double x,y,xmin,xmax,ymin,ymax;
   int i;

   xmin = xmax = locx;
   ymin = ymax = locy;   /* Find the bounding box */
   for (i=0; i<(vertices - 1); ++i)
      { if ((x = (locx+vx[i])) > xmax) xmax = x;
	if ((x = (locx+vx[i])) < xmin) xmin = x;
	if ((y = (locy+vy[i])) > ymax) ymax = y;
	if ((y = (locy+vy[i])) < ymin) ymin = y;
      }
   *lx = xmin; *rx = xmax; *ty = ymax; *by = ymin;
}




void
SplineImpl::transSpecial(char *type1, double dx, double dy)
{
   int i,v;

   if (streql(type1,"FILL"))
      { width +=  dx;
	if (width < 0.0)
	   width = 0.0;
	else if (width > 1.0)
	   width = 1.0;
        style +=  dy;
	if (style < 0.0)
	   style = 0.0;
	else if (style > 1.0)
	   style = 1.0;
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
   else if (streqln(type1,"RESIZE",6)) {
      if (('1' <= *(type1+6)) && (*(type1+6) <= '7')) {
         DamageIt();
         v = (int) (*(type1+6) - '0');
         for (i=v-1; i<vertices-1; i++) {
            vx[i] += dx;
            vy[i] += dy;
            }
         DamageIt();
         }
      }      
   else if (streqln(type1,"GRAB",4)) {
      if (('1' <= *(type1+4)) && (*(type1+4) <= '7')) {
         DamageIt();
         v = (int) (*(type1+4) - '1');
         vx[v] += dx;
         vy[v] += dy;
         DamageIt();
         }
      }      
}

   


void
SplineImpl::Draw()
{
   drawer(color);
}

void
SplineImpl::Erase()
{
   drawer(view->GetBgColor());
}


void
SplineImpl::drawer(COLORINDEX col)
{
   int i,x[32],y[32];
   LINE_STYLE lstyle,old = 0;

   if (!visibility)
      return;

   x[0] = view->TO_PIX_X(locx);
   y[0] = view->TO_PIX_Y(locy);
   for (i=1; i<vertices; i++) {
      x[i] = view->TO_PIX_X(locx + vx[i-1]);
      y[i] = view->TO_PIX_Y(locy + vy[i-1]);
      }
   set_color(col);
   lstyle = get_linestyle(width,style);
   if (lstyle != SOLID)
      old = line_style(lstyle);
   make_a_spline(vertices,x,y,view);
   if (lstyle != SOLID)
      line_style(old);
}




Polygon& 
Polygon::operator=(const Polygon& rhs) 
{
   if  (this == &rhs) return *this;
   AnimObject::operator=(rhs);
   return *this;
}




PolygonImpl::PolygonImpl(View *a, int v, double lx, double ly, int ve, 
                 double vtx[], double vty[], COLOR c, double f) 
                 : AnimObjectImpl(a,v,lx,ly)
{
   int i;

   type = P_Polygon;
   vertices = ve;
   for (i=0; i<vertices-1; i++) {
       vx[i] = vtx[i];
       vy[i] = vty[i];
     }
   strcpy(colorname,c);
   color = load_color(colorname);
   fill = f;
}



PolygonImpl::PolygonImpl(const PolygonImpl& p) : AnimObjectImpl( p )
{
   vertices = p.vertices;
   for (int i=0; i<vertices-1; i++) {
       vx[i] = p.vx[i];
       vy[i] = p.vy[i];
     }
   strcpy(colorname,p.colorname);
   color = p.color;
   fill = p.fill;
}



PolygonImpl::~PolygonImpl()
{
   Erase();
}


void 
PolygonImpl::GetValues(View* *vi, int *v, double *lx, double *ly, int *ve, 
                double vtx[], double vty[], COLOR c, double *f)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (ve) *ve=vertices;
   for (int i=0; i<vertices; i++) {
       if (vtx) vtx[i] = vx[i];
       if (vty) vty[i] = vy[i];
   }
   if (c) strcpy(c,colorname);
   if (f) *f=fill;
}	


LocPtr
PolygonImpl::Where(PART p)
{
   double	   xmin,ymin,xmax,ymax;

   BoundBox(&xmin,&ymin,&xmax,&ymax);

   switch (p)
   {
      case PART_C :
	 return( new Loc((xmin + xmax)/2.0,(ymin + ymax)/2.0) );
      case PART_NW :
	 return( new Loc(xmin,ymax) );
      case PART_N :
	 return( new Loc((xmin+ xmax)/2.0,ymax) );
      case PART_NE :
	 return( new Loc(xmax,ymax) );
      case PART_E :
	 return( new Loc(xmax,(ymin + ymax)/2.0) );
      case PART_SE :
	 return( new Loc(xmax,ymin) );
      case PART_S :
	 return( new Loc((xmin + xmax)/2.0,ymin) );
      case PART_SW :
	 return( new Loc(xmin,ymin) );
      case PART_W :
	 return( new Loc(xmin,(ymin + ymax)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}



void
PolygonImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   double x,y,xmin,xmax,ymin,ymax;
   int i;

   xmin = xmax = locx;
   ymin = ymax = locy;   /* Find the bounding box */
   for (i=0; i<(vertices - 1); ++i)
      { if ((x = (locx+vx[i])) > xmax) xmax = x;
	if ((x = (locx+vx[i])) < xmin) xmin = x;
	if ((y = (locy+vy[i])) > ymax) ymax = y;
	if ((y = (locy+vy[i])) < ymin) ymin = y;
      }
   *lx = xmin; *rx = xmax; *ty = ymax; *by = ymin;
}




void
PolygonImpl::transSpecial(char *type1, double dx, double dy)
{
   int i,v;

   if (streql(type1,"FILL"))
      { fill +=  dx;
	if (fill < 0.0)
	   fill = 0.0;
	else if (fill > 1.0)
	   fill = 1.0;
        DamageIt();
      }
   else if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
   else if (streqln(type1,"RESIZE",6)) {
      if (('1' <= *(type1+6)) && (*(type1+6) <= '7')) {
         DamageIt();
         v = (int) (*(type1+6) - '0');
         for (i=v-1; i<vertices-1; i++) {
            vx[i] += dx;
            vy[i] += dy;
            }
         DamageIt();
         }
      }      
   else if (streqln(type1,"GRAB",4)) {
      if (('1' <= *(type1+4)) && (*(type1+4) <= '7')) {
         DamageIt();
         v = (int) (*(type1+4) - '1');
         vx[v] += dx;
         vy[v] += dy;
         DamageIt();
         }
      }      
}

   


void
PolygonImpl::Draw()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   drawer(color,curfill);
}

void
PolygonImpl::Erase()
{
   FILL_STYLE curfill;

   curfill = get_fillstyle(fill);
   if (curfill == POLKA_FILL_OUTLINE)
      drawer(view->GetBgColor(),POLKA_FILL_OUTLINE);
   else
      drawer(view->GetBgColor(),POLKA_FILL_SOLID);
}


void
PolygonImpl::drawer(COLORINDEX col, FILL_STYLE curfill)
{
   int i;
   FILL_STYLE oldfill;
   XPoint pts[10];

   if (!visibility)
      return;

   pts[0].x = view->TO_PIX_X(locx);
   pts[0].y = view->TO_PIX_Y(locy);
   for (i=1; i<vertices; i++) {
      pts[i].x = view->TO_PIX_X(locx + vx[i-1]);
      pts[i].y = view->TO_PIX_Y(locy + vy[i-1]);
      }
   set_color(col);
   if (curfill == POLKA_FILL_OUTLINE)
      {
        pts[vertices].x = view->TO_PIX_X(locx);
        pts[vertices].y = view->TO_PIX_Y(locy);
        XDrawLines(_display,view->DrawWindow(),inq_gc(),pts,
                               vertices+1,CoordModeOrigin);
      }
   else
      {
        oldfill = fill_style(curfill);
        XFillPolygon(_display,view->DrawWindow(),inq_gc(),pts,
                             vertices,Nonconvex,CoordModeOrigin);
        fill_style(oldfill);
      }
}



Text& 
Text::operator=(const Text& rhs) 
{
   if  (this == &rhs) return *this;
   AnimObject::operator=(rhs);
   return *this;
}



TextImpl::TextImpl(View *a, int v, double lx, double ly, COLOR c, 
		   const char *fn, const char *ts, int o) : AnimObjectImpl(a,v,lx,ly)
{
   type = P_Text;
   strcpy(colorname,c);
   color = load_color(colorname);
   if (fn) { 
      fontid = load_font((char*)fn); 
      strcpy(fontname,fn);
      }
   else {
      fontid = inq_font(); 
      fontname[0] = '\0';
      }
   strcpy(tstring,(char*)ts);
   orient = o;
   text_info(fontid,(char*)ts,&xext,&yext,&xoff,&yoff);
}


TextImpl::TextImpl(const TextImpl& t) : AnimObjectImpl( t )
{
   strcpy(colorname,t.colorname);
   color = t.color;
   fontid = t.fontid; 
   strcpy(fontname,t.fontname);
   strcpy(tstring,t.tstring);
   orient = t.orient;
   xext = t.xext;
   yext = t.yext;
   xoff = t.xoff;
   yoff = t.yoff;
}
   


TextImpl::~TextImpl()
{
   Erase();
}


void 
TextImpl::GetValues(View* *vi, int *v, double *lx, double *ly, COLOR c, 
            char* fn, char* ts, int *o)
{
   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (c) strcpy(c,colorname);
   if (fn) strcpy(fn,fontname);
   if (ts) strcpy(ts,tstring);
   if (o) *o=orient;
}	


LocPtr
TextImpl::Where(PART p)
{
   double	   lx,by,rx,ty;

   BoundBox(&lx,&by,&rx,&ty);

   switch (p)
   {
      case PART_C :
	 return( new Loc((lx + rx)/2.0,(by + ty)/2.0) );
      case PART_NW :
	 return( new Loc(lx,ty) );
      case PART_N :
	 return( new Loc((lx+ rx)/2.0,ty) );
      case PART_NE :
	 return( new Loc(rx,ty) );
      case PART_E :
	 return( new Loc(rx,(ty + by)/2.0) );
      case PART_SE :
	 return( new Loc(rx,by) );
      case PART_S :
	 return( new Loc((lx + rx)/2.0,by) );
      case PART_SW :
	 return( new Loc(lx,by) );
      case PART_W :
	 return( new Loc(lx,(by + ty)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}



void
TextImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   double ex,ox,ey,oy;

   ex = (double) xext / (double) view->PIXX();
   ox = (double) xoff / (double) view->PIXX();
   ey = (double) yext / (double) view->PIXY();
   oy = (double) yoff / (double) view->PIXY();
   if (orient == 0) /* lower left */
      { *lx = locx - ox;
	*rx = locx + ex;
	*by = locy - oy;
	*ty = locy + ey;
      }
   else
      { *lx = locx - ex/2.0;
	*rx = locx + ex/2.0;
	*by = locy - ey/2.0;
	*ty = locy + ey/2.0 + oy; /*seems to need that little extra */
      }
}




void
TextImpl::transSpecial(char *type1, double dx, double dy)
{
   if (streql(type1,"COLOR")) {
      color = get_pathcolor(dx,dy);
      DamageIt();
      }
}

   


void
TextImpl::Draw()
{
   drawer(color);
}

void
TextImpl::Erase()
{
   drawer(view->GetBgColor());
}


void
TextImpl::drawer(COLORINDEX col)
{
   double x0,y0;
   int oldfont; 
   FILL_STYLE oldfill;

   if (!visibility)
      return;

   if (orient == 0)
      {
        x0 = locx;
        y0 = locy;
      }
   else
      {
        x0 = locx - ((xext/2.0 - xoff) / view->PIXX());
        y0 = locy - ((yext/2.0 - yoff) / view->PIXY());
      }

   set_color(col);
   oldfont = set_font(fontid);
   oldfill = fill_style(POLKA_FILL_SOLID);
   XDrawString(_display,view->DrawWindow(),inq_gc(),
                view->TO_PIX_X(x0),view->TO_PIX_Y(y0),tstring,strlen(tstring));
   fill_style(oldfill);
   set_font(oldfont);   
}




Bitmap& 
Bitmap::operator=(const Bitmap& rhs) 
{
   if  (this == &rhs) return *this;
   AnimObject::operator=(rhs);
   return *this;
}




BitmapImpl::BitmapImpl(View *vi, int v, double lx, double ly, int height, 
         int width, char data[], COLOR fg, COLOR bg) 
          : AnimObjectImpl(vi,v,lx,ly)
{
   FILL_STYLE oldfill;
   COLORINDEX fgcolor,bgcolor;
   Pixel xfg,xbg;
   int bmsize,i;

   type = P_Bitmap;
   iheight = height;
   iwidth = width;
   strcpy(fgcolname,fg);
   strcpy(bgcolname,bg);
   bmsize = (width / 4) * height;   // char is of size=4
   bmdata = new char[bmsize];
   for (i=0; i<bmsize; i++)
      *(bmdata+i) = data[i];
   /* Must remove clip_rectangle set by animation loop damage routines */
   XSetClipMask(_display, inq_gc(), None);
   if (_color_screen) {
     fgcolor = load_color(fg);
     bgcolor = load_color(bg);
     xfg = GetXColor(fgcolor);
     xbg = GetXColor(bgcolor);
     if ((pixmap = XCreatePixmapFromBitmapData(_display,view->DrawWindow(),
                      bmdata, width, height, xfg, xbg, 
                      DefaultDepth(_display,DefaultScreen(_display))))
                == None) {
         POLKA_complain("BitmapImpl::Bitmap","Bad Pixmap--Creating blank default\n");
         pixmap = XCreatePixmap(_display, view->DrawWindow(), width,
                   height, DefaultDepth(_display,DefaultScreen(_display)));
         set_color(WHITE);
         oldfill = fill_style(40);
         XFillRectangle(_display,view->DrawWindow(),inq_gc(), 0, 0,
            width,height);
         fill_style(oldfill);
         }
     }
   else {
      if ((pixmap = XCreateBitmapFromData(_display,view->DrawWindow(),
                      bmdata, width, height)) == None) {
         POLKA_complain("BitmapImpl::Bitmap","Bad Pixmap--Creating blank default\n");
         pixmap = XCreatePixmap(_display, view->DrawWindow(), width,
                   height, DefaultDepth(_display,DefaultScreen(_display)));
         set_color(WHITE);
         oldfill = fill_style(40);
         XFillRectangle(_display,view->DrawWindow(),inq_gc(), 0, 0,
            width,height);
         fill_style(oldfill);
         }
     }

      
}


BitmapImpl::BitmapImpl(const BitmapImpl& b) : AnimObjectImpl( b )
{
   int bmsize,i;

   iheight = b.iheight;
   iwidth = b.iwidth;
   strcpy(fgcolname,b.fgcolname);
   strcpy(bgcolname,b.bgcolname);
   bmsize = (iwidth / 4) * iheight;   // char is of size=4
   bmdata = new char[bmsize];
   for (i=0; i<bmsize; i++)
      *(bmdata+i) = b.bmdata[i];
   /* Must remove clip_rectangle set by animation loop damage routines */
   XSetClipMask(_display, inq_gc(), None);
   pixmap = XCreatePixmap(_display, view->DrawWindow(), b.iwidth,
              b.iheight, DefaultDepth(_display,DefaultScreen(_display)));
   XCopyArea(_display,b.pixmap,pixmap,inq_gc(),0,0,
               b.iwidth,b.iheight,0,0);
}
   


BitmapImpl::~BitmapImpl()
{
   Erase();
   delete(bmdata);
   XFreePixmap(_display,pixmap);
}


void 
BitmapImpl::GetValues(View* *vi, int *v, double *lx, double *ly, 
            int *wid, int *hei, char data[], COLOR fg, COLOR bg)
{
   int bmsize,i;

   if (vi) *vi=view;
   if (v) *v=visibility;
   if (lx) *lx=locx;
   if (ly) *ly=locy;
   if (wid) *wid=iwidth;
   if (hei) *hei=iheight;
   if (data) {
      bmsize = (iwidth / 4) * iheight;   // char is of size=4
      for (i=0; i<bmsize; i++)
         data[i] = bmdata[i];
      }
   if (fg) strcpy(fg,fgcolname);
   if (bg) strcpy(bg,bgcolname);
}	


LocPtr
BitmapImpl::Where(PART p)
{
   double	   lx,by,rx,ty;

   BoundBox(&lx,&by,&rx,&ty);

   switch (p)
   {
      case PART_C :
	 return( new Loc((lx + rx)/2.0,(by + ty)/2.0) );
      case PART_NW :
	 return( new Loc(lx,ty) );
      case PART_N :
	 return( new Loc((lx+ rx)/2.0,ty) );
      case PART_NE :
	 return( new Loc(rx,ty) );
      case PART_E :
	 return( new Loc(rx,(ty + by)/2.0) );
      case PART_SE :
	 return( new Loc(rx,by) );
      case PART_S :
	 return( new Loc((lx + rx)/2.0,by) );
      case PART_SW :
	 return( new Loc(lx,by) );
      case PART_W :
	 return( new Loc(lx,(by + ty)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}



void
BitmapImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   *lx = locx;
   *by = locy;
   *rx = locx + view->SIZE_REAL_X(iwidth);
   *ty = locy + view->SIZE_REAL_Y(iheight);
}




void
BitmapImpl::Draw()
{
   if (!visibility)
      return;
   XCopyArea(_display,pixmap,view->DrawWindow(),inq_gc(),0,0,
               iwidth,iheight,view->TO_PIX_X(locx),
                              view->TO_PIX_Y(locy)-iheight);
}

void
BitmapImpl::Erase()
{
   FILL_STYLE oldfill;

   if (!visibility)
      return;

   set_color(view->GetBgColor());
   oldfill = fill_style(40);
   XFillRectangle(_display,view->DrawWindow(),inq_gc(),
          view->TO_PIX_X(locx),view->TO_PIX_Y(locy)-iheight,iwidth,iheight);
   fill_style(oldfill);
}


void
BitmapImpl::transSpecial(char *, double, double)
{
}







Set& 
Set::operator=(const Set& rhs) 
{
   if  (this == &rhs) return *this;
   AnimObject::operator=(rhs);
   return *this;
}





SetImpl::SetImpl(View *vi) : AnimObjectImpl(vi, 1, 0.0, 0.0)
{
   type = P_Set;
}



LocPtr
SetImpl::Where(PART p)
{
   double	   lx,by,rx,ty;

   BoundBox(&lx,&by,&rx,&ty);

   switch (p)
   {
      case PART_C :
	 return( new Loc((lx + rx)/2.0,(by + ty)/2.0) );
      case PART_NW :
	 return( new Loc(lx,ty) );
      case PART_N :
	 return( new Loc((lx+ rx)/2.0,ty) );
      case PART_NE :
	 return( new Loc(rx,ty) );
      case PART_E :
	 return( new Loc(rx,(ty + by)/2.0) );
      case PART_SE :
	 return( new Loc(rx,by) );
      case PART_S :
	 return( new Loc((lx + rx)/2.0,by) );
      case PART_SW :
	 return( new Loc(lx,by) );
      case PART_W :
	 return( new Loc(lx,(by + ty)/2.0) );
   }
   return( new Loc(0.0,0.0) );
}


void 
SetImpl::BoundBox(double *lx, double *by, double *rx, double *ty)
{
   double plx,pby,prx,pty;
   Set *s;
   AOPtr objs;
   
   s = (Set *)wrapper;
   objs = s->Objs();
   if (!objs)
     { *lx = *by = *rx = *ty = 0.0;
         return;
     }
   objs->object->BoundBox(lx,by,rx,ty);
   
   for (AOPtr a = objs->next; a; a=a->next)
       { a->object->BoundBox(&plx,&pby,&prx,&pty);
         if (plx < *lx) *lx = plx;
         if (prx > *rx) *rx = prx;
         if (pby < *by) *by = pby;
         if (pty > *ty) *ty = pty;
     }
}


void
SetImpl::Draw()
{
   Set *s = (Set *) wrapper;
   for (AOPtr a=s->Objs(); a; a=a->next)
      a->object->Draw();
}


void
SetImpl::Erase()
{
   Set *s = (Set *) wrapper;
   for (AOPtr a=s->Objs(); a; a=a->next)
      a->object->Erase();
}
  





COLORINDEX
get_pathcolor(double dx, double)
{
   COLOR_MAPPER *cl;

   for (cl = _color_list; cl; cl=cl->next)
       if (ABS(dx - cl->pathval) < .0001) // roughly equal
          return(cl->colornum);
   return(BLACK);      
}





/***************************************************************/
/*							       */
/*   get_color - given a string of a POLKA_COLOR, return its   */
/*	integer value.					       */
/*							       */
/***************************************************************/

COLORINDEX
get_color(COLOR str)
{
   if (strcmp("WHITE",str) == 0)
      return( WHITE );
   else if (strcmp("BLACK",str) == 0)
      return( BLACK );
   else if (strcmp("RED",str) == 0)
      return( RED );
   else if (strcmp("ORANGE",str) == 0)
      return( ORANGE );
   else if (strcmp("YELLOW",str) == 0)
      return( YELLOW );
   else if (strcmp("GREEN",str) == 0)
      return( GREEN );
   else if (strcmp("BLUE",str) == 0)
      return( BLUE );
   else if (strcmp("MAROON",str) == 0)
      return( MAROON );
   else
      return( BLACK );
}



/***************************************************************/
/*							       */
/*   get_fillstyle - given a fill value between 0.0 and 1.0    */
/*	return the POLKA fill style it corresponds to.	       */
/*							       */
/***************************************************************/

FILL_STYLE
get_fillstyle(double val)
{
//  return (val >= 0.0 && val <= 1.0 ? irint(val * 40.0) : 40); /* 40 fill */
							      /* styles */

      return ((val >= 0.0) && (val <= 1.0) ? int(val * 40.0 + 0.5) : 40);
}



/***************************************************************/
/*							       */
/*   get_linestyle - given a width and a style, return an      */
/*	LINE_STYLE. 	         			       */
/*							       */
/***************************************************************/

LINE_STYLE
get_linestyle(double width,double style)
{
   if (width > 0.666667)
      { if (style < 0.333333)
	   return(THICKER_DOTTED);
	else if (style < 0.666667)
	   return(THICKER_DASHED);
	else
	   return(THICKER);
      }
   else if (width > 0.333333)
      { if (style < 0.333333)
	   return(THICK_DOTTED);
	else if (style < 0.666667)
	   return(THICK_DASHED);
	else
	   return(THICK);
      }
   else
      { if (style < 0.333333)
	   return(DOTTED);
	else if (style < 0.666667)
	   return(DASHED);
	else
	   return(SOLID);
      }
}





/*  #########################################################  */
/*   compute_arrow_off - figure out what the offsets should    */
/*	be for the points of an arrow.			       */
/*	I haven't got a clue how this works, but it just does. */
/*	Many thanks to Eric Golin for code.		       */
/*  #########################################################  */

#define   ARC_HEAD_LENGTH   ((double)0.025)
#define   ARC_HEAD_ANGLE    MYPI/2.5

void
compute_arrow_off(double dx,double dy,double *adx1,double *ady1,
                  double *adx2,double *ady2)
{
   double arclen,len,theta,alpha,beta;

   if ((dx == 0.0) && (dy == 0.0))
      { *adx1 = *adx2 = *ady1 = *ady2 = 0.0;
	return;
      }

   arclen = ARC_HEAD_LENGTH;
   len = sqrt((dx*dx) + (dy*dy));
   if (len < arclen)
      arclen = len;

   theta = atan2(dx,dy);
   alpha = -(theta + ARC_HEAD_ANGLE);
   beta = MYPI - (theta - ARC_HEAD_ANGLE);
   *adx1 = arclen * cos(alpha);
   *ady1 = arclen * sin(alpha);   /* minus, because +y goes up */
   *adx2 = arclen * cos(beta);
   *ady2 = arclen * sin(beta);
}



/*  #########################################################  */
/*   arrow_poly_direction - determine what direction to        */
/*	point the arrowheads on a polyline.  Tricky because    */
/*	offsets at the two endpoints may be 0,0.  We want the  */
/*	last non-negative pair. 			       */
/*  #########################################################  */

void
arrow_poly_direction(int num,double vertx[7],double verty[7],
                 double *fx,double *fy,double *bx,double *by)
{
   double    tempx[8],tempy[8];
   int	     i;

   for (i=1; i<num; ++i)   /* need a starting position too */
      { tempx[i] = vertx[i-1];
	tempy[i] = verty[i-1];
      }
   tempx[0] = tempy[0] = 0.0;

   for (i=num-1; i>=1; --i)
      { *fx = tempx[i] - tempx[i-1];
	*fy = tempy[i] - tempy[i-1];
	if ((*fx != 0.0) || (*fy != 0.0))
	   break;
      }

   for (i=1; i<num; ++i)
      { *bx = tempx[i-1] - tempx[i];
	*by = tempy[i-1] - tempy[i];
	if ((*bx != 0.0) || (*by != 0.0))
	   break;
      }
}



