/**********************************************************************
 **                                                                  **
 **                           mission_expert.h                       **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  Copyright 2002 - 2005 Georgia Tech Research Corporation         **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: mission_expert.h,v 1.1.1.1 2006/07/12 13:37:54 endo Exp $ */

#ifndef MISSION_EXPERT_H
#define MISSION_EXPERT_H

#include <stdio.h>
#include <time.h>
#include <string>
#include <vector>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <Xm/Xm.h>
#include <pthread.h>

#include "symbol.hpp"
#include "load_rc.h"
#include "mission_expert_types.h"
#include "configuration.hpp"
#include "cbrwizard.h"

using std::string;
using std::vector;

class CfgEditCBRClient;

// Extentions of some of the output files
const string MEXP_EXTENSION_FEATUREFILENAME = "data";
const string MEXP_EXTENSION_FEATURESTATUSFILENAME = "status";
const string MEXP_EXTENSION_CDL = "cdl";
const string MEXP_EXTENSION_ACDL = "acdl";
const string MEXP_EXTENSION_BACKUP = "backup";

// Range of the slider bar
const string MEXP_DEFAULT_SLIDER1_RANAGE_TITLE = "[0.01 - 1.00]";
const string MEXP_DEFAULT_SLIDER10_RANAGE_TITLE = "[1 - 10]";
const string MEXP_DEFAULT_SLIDER100_RANAGE_TITLE = "[1 - 100]";

// Some strings
const string MEXP_STRING_OVERLAY = "overlay";
const string MEXP_STRING_TASK = "task";
const string MEXP_STRING_COORDINATE = "coordinate";
const string MEXP_STRING_FEATURE = "feature";
const string MEXP_STRING_OPTION = "option";
const string MEXP_STRING_CANCEL = "cancel";
const string MEXP_STRING_CONFIGURE = "configure";
const string MEXP_STRING_REPAIR = "repair";
const string MEXP_STRING_REPAIR_FAILURE = "repair-failure";
const string MEXP_STRING_YES = "Yes";
const string MEXP_STRING_NO = "No";
const string MEXP_STRING_RETRIEVE = "retrieve";
const string MEXP_STRING_ADD = "add";
const string MEXP_STRING_GLOBALFEATURE = "global-feature";
const string MEXP_STRING_VIA_POINT = "via-point";
const string MEXP_STRING_ROBOT = "Robot";
const string MEXP_STRING_INFINITY = "inf";
const string MEXP_STRING_TASK_CONSTRAINTS = "task-constraints";
const string MEXP_STRING_TASK_CONSTRAINT = "task-constraint";
const string MEXP_STRING_ROBOT_CONSTRAINTS = "robot-constraints";
const string MEXP_STRING_ROBOT_CONSTRAINT = "robot-constraint";
const string MEXP_STRING_CNP = "cnp";

// RC file name.
const string MEXP_RC_FILENAME = ".mission_expertrc";

// RC file vocabulary
const string MEXP_RC_STRING_FEATURE_KEY = "<FEATURE";
const string MEXP_RC_STRING_FEATURE_BEGIN = "BEGIN>";
const string MEXP_RC_STRING_FEATURE_END = "END>";
const string MEXP_RC_STRING_FEATURE_TYPE = "FEATURE-TYPE";
const string MEXP_RC_STRING_FEATURE_NAME = "FEATURE-NAME";
const string MEXP_RC_STRING_FEATURE_NON_INDEX = "FEATURE-NON-INDEX";
const string MEXP_RC_STRING_FEATURE_HIDE = "FEATURE-HIDE";
const string MEXP_RC_STRING_FEATURE_WEIGHT = "FEATURE-WEIGHT";
const string MEXP_RC_STRING_FEATURE_OPTION_TYPE = "FEATURE-OPTION-TYPE";
const string MEXP_RC_STRING_FEATURE_SELECTED_OPTION = "FEATURE-SELECTED-OPTION";
const string MEXP_RC_STRING_FEATURE_OPTION = "FEATURE-OPTION";
const string MEXP_RC_STRING_TASK_KEY = "<TASK";
const string MEXP_RC_STRING_TASK_BEGIN = "BEGIN>";
const string MEXP_RC_STRING_TASK_END = "END>";
const string MEXP_RC_STRING_TASK_NAME = "TASK-NAME";
const string MEXP_RC_STRING_TASK_BUTTON_NAME = "TASK-BUTTON-NAME";
const string MEXP_RC_STRING_TASK_SPECIFIC_SELECTED_OPTION = "TASK-SPECIFIC-SELECTED-OPTION";
const string MEXP_RC_STRING_TASK_MAX_USAGE = "TASK-MAX-USAGE";
const string MEXP_RC_STRING_TASK_RUNTIME_CNP = "TASK-RUNTIME-CNP";
const string MEXP_RC_STRING_DEFAULT_TASK = "DEFAULT-TASK";
const string MEXP_RC_STRING_DEFAULT_TASK_WEIGHT = "DEFAULT-WEIGHT";
const string MEXP_RC_STRING_RUTIME_CNP_MISSION_MANAGER_TASK = "RUTIME-CNP-MISSION-MANAGER-TASK";
const string MEXP_RC_STRING_RUTIME_CNP_BIDDER_TASK = "RUTIME-CNP-BIDDER-TASK";
const string MEXP_RC_STRING_GLOBAL = "GLOBAL";
const string MEXP_RC_STRING_LOCAL = "LOCAL";
const string MEXP_RC_STRING_TOGGLE = "TOGGLE";
const string MEXP_RC_STRING_SLIDER = "SLIDER";
const string MEXP_RC_STRING_CONSTRAINT_KEY = "<CONSTRAINT";
const string MEXP_RC_STRING_CONSTRAINT_BEGIN = "BEGIN>";
const string MEXP_RC_STRING_CONSTRAINT_END = "END>";
const string MEXP_RC_STRING_CONSTRAINT_TYPE = "CONSTRAINT-TYPE";
const string MEXP_RC_STRING_CONSTRAINT_NAME = "CONSTRAINT-NAME";
const string MEXP_RC_STRING_CONSTRAINT_HIDE = "CONSTRAINT-HIDE";
const string MEXP_RC_STRING_CONSTRAINT_OPTION_TYPE = "CONSTRAINT-OPTION-TYPE";
const string MEXP_RC_STRING_CONSTRAINT_SELECTED_OPTION = "CONSTRAINT-SELECTED-OPTION";
const string MEXP_RC_STRING_CONSTRAINT_OPTION = "CONSTRAINT-OPTION";
const string MEXP_RC_STRING_ROBOT_CONSTRAINTS_KEY = "<ROBOT-CONSTRAINTS";
const string MEXP_RC_STRING_ROBOT_CONSTRAINTS_BEGIN = "BEGIN>";
const string MEXP_RC_STRING_ROBOT_CONSTRAINTS_END = "END>";
const string MEXP_RC_STRING_ROBOT_CONSTRAINTS_ROBOT_ID = "ROBOT-CONSTRAINTS-ROBOT-ID";

// Misc setups.
const int MEXP_USAGE_INFINITY = 0;

class MissionExpert : public CBRWizard {

public:
    typedef enum FeatureStatus_t {
        CANCEL,
        CONFIGURE,
        REPAIR,
        REPAIR_FAILURE
    };

protected:
    typedef enum StatusWindowNavigation_t {
        STATUS_WINDOW_PREV,
        STATUS_WINDOW_NEXT
    };

    typedef enum ThreadedXCommandsTypes_t {
        X_COMMAND_UPDATE_STATUS_WINDOW,
        X_COMMAND_HIGHLIGHT_STATES
    };

    typedef enum RevertStatus_t {
        REVERT_FAILURE,
        REVERT_SUCCESS,
        REVERT_NO_ORIGINAL_FILE,
        REVERT_NO_BACKUP_FILE
    };

    typedef struct FeatureMenuCallbackData_t {
        int featureId;
        int optionType;
        float optionValue;
        MissionExpert *missionExpertInstance;
    };

    typedef struct GlobalFeatureMenuCallbackData_t {
        int globalFeaturesId;
        int optionType;
        float optionValue;
        MissionExpert *missionExpertInstance;
    };

    typedef struct FeatureMenuTaskCallbackData_t {
        int taskId;
        MissionExpert *missionExpertInstance;
    };

    typedef struct StatusWindows_t {
        Widget popupShell;
        Widget titleBox;
        Widget textBox;
        Widget prevBtn;
        Widget nextBtn;
        Widget loadBtn;
        Widget cancelBtn;
        Widget ratingBox;
        Widget *starBox;
        int textPos;
        bool isUp;
        int currentMissionId;
        int selectedMissionId;
    };

    typedef struct XCommandUpdateStatesWindow_t {
        int rating;
        string missionSummary;
        bool prevBtnIsSensitive;
        bool nextBtnIsSensitive;
        bool loadBtnIsSensitive;
        bool cancelBtnIsSensitive;
        int currentMissionId;
    };

    typedef struct XCommandHighlightStates_t {
        vector<MExpStateInfo_t> stateInfoList;
    };

    typedef struct PendingThreadedXCommands_t {
        int type;
        int index;
    };

    typedef struct CheckForkedMlabStatusData_t {
        MissionExpert *missionExpertInstance;
        int pid;
        bool eventLogIsON;
    };

protected:
    FILE *featureStatusFile_;
    FILE *resultFile_;
    Widget featureMenu_w_;
    Widget **featureMenuOption_w_;
    Widget *featureMenuTask_w_;
    Widget **globalFeaturesMenuOption_w_;
    Widget *globalFeaturesMenuTask_w_;
    StatusWindows_t statusWindows_;
    Pixmap starPixmap_;
    Pixmap emptyStarPixmap_;
    MExpMissionTask_t featureMenuUpTask_;
    MExpRetrievedMissionSummary_t retrievedMissionSummary_;
    MExpStateInfo_t lastHighlightedStateInfo_;
    MExpRobotIDs_t currentRobotIDs_;
    CfgEditCBRClient *cfgeditCBRClient_;
    vector<MExpFeature_t> featureMenuUpGlobalFeatures_;
    vector<MExpFeature_t> fixedGlobalFeatures_;
    vector<MExpLogfileInfo_t> logfileInfoList_;
    vector<XCommandUpdateStatesWindow_t> shXCommandUpdateStatesWindowList_;
    vector<XCommandHighlightStates_t> shXCommandHighlightStatesList_;
    vector<PendingThreadedXCommands_t> shPendingThreadedXCommands_;
    vector<string> createdDataFileList_;
    vector<int> currentCBRLibraryDataIndexList_;
    string featureFileString_;
    string resultFileName_;
    string missionPlanFileName_;
    string robotName_;
    string dataFilesDirectory_;
    int mexpRunMode_;
    bool disableMExp_;
    bool dataFilesNamed_;
    bool keepDataFiles_;
    bool featuresLoaded_;
    bool featureMenuIsUp_;
    bool shouldSaveMissionToCBRLibrary_;
    bool missionSavedToCBRLibrary_;
    bool missionLoadedFromCBRLibrary_;
    bool disableRepair_;

    const MExpStateInfo_t EMPTY_STATE_INFO_; 

    static pthread_mutex_t threadedXCommandsMutex_;

    static const string RCTABLE_DISABLE_STRING_;
    static const string RCTABLE_DISABLE_REPAIR_STRING_;
    static const string RCTABLE_DISABLE_PREMISSION_CNP_STRING_;
    static const string RCTABLE_DISABLE_RUNTIME_CNP_STRING_;
    static const string RCTABLE_DATAFILESDIR_STRING_;
    static const string DEFAULT_FEATUREMENU_TITLE_;
    static const string DEFAULT_FEATUREMENU_TASK_LABEL_;
    static const string DEFAULT_FEATUREMENU_SAVE_BUTTON_LABEL_;
    static const string DEFAULT_FEATUREMENU_CANCEL_BUTTON_LABEL_;
    static const string DEFAULT_STATUSWINDOW_TITLE_;
    static const string DEFAULT_STATUSWINDOW_RATING_LABEL_;
    static const string DEFAULT_STATUSWINDOW_NAVIGATION_LABEL_;
    static const string DEFAULT_STATUSWINDOW_PREV_BUTTON_LABEL_;
    static const string DEFAULT_STATUSWINDOW_NEXT_BUTTON_LABEL_;
    static const string DEFAULT_STATUSWINDOW_LOAD_BUTTON_LABEL_;
    static const string DEFAULT_STATUSWINDOW_CANCEL_BUTTON_LABEL_;
    static const string ACDL_STRING_AGENT_NAME_;
    static const string ACDL_STRING_TASK_INDEX_;
    static const string ACDL_STRING_START_;
    static const string CDL_STRING_MAXVELOCITY_;
    static const string CDL_STRING_BASEVELOCITY_;
    static const string CDL_STRING_START_SUB_MISSION_;
    static const string FEATURE_NAME_NUMROBOTS_;
    static const string FEATURE_NAME_MAXVELOCITY_;
    static const string FEATURE_NAME_AGGRESSIVENESS_;
    static const string CONFIRM_SUCCESSFUL_MISSION_PHRASE_;
    static const string CONFIRM_VIEW_ANYWAY_PHRASE_;
    static const string ASK_USE_MEXP_PHRASE_;
    static const string ASK_DIAGNOSIS_PHRASE_;
    static const string ASK_CNP_PHRASE_;
    static const string ASK_CNP_PHRASE_SIMPLE_;
    static const string ASK_PREMISSION_CNP_PHRASE_;
    static const string ASK_PREMISSION_CNP_PHRASE_SIMPLE_;
    static const string ASK_RUNTIME_CNP_PHRASE_;
    static const string ASK_RUNTIME_CNP_PHRASE_SIMPLE_;
    static const string REPAIR_NO_SOLUTION_PHRASE_;
    static const string REPAIR_SOLUTION_PHRASE_;
    static const string FSASUMMARY_EMPTY_PHRASE_;
    static const string FSASUMMARY_DEFAULT_PHRASE_;
    static const double REPLAY_DISPLAY_RATIO_;
    static const int DEFAULT_FEATUREMENU_WIDTH_;
    static const int DEFAULT_FEATUREMENU_LINEHEIGHT_;
    static const int DEFAULT_STATUSWINDOW_WIDTH_;
    static const int DEFAULT_STATUSWINDOW_LINEHEIGHT_;
    static const int DEFAULT_STATUSWINDOW_TOPTITLE_HEIGHT_;
    static const int DEFAULT_STATUSWINDOW_TEXT_HEIGHT_;
    static const int DEFAULT_STATUSWINDOW_TEXT_NUMROWS_;
    static const int MAX_SUITABILITY_RATING_;
    static const int EVENT_WAITER_SLEEP_TIME_USEC_;
    static const int PROCESS_THREADED_X_COMMAND_TIMER_MSEC_;
    static const bool SEND_FEEDBACKS_TO_LIBRARY_;
    static const bool INCREMENTALLY_UPDATE_LIBRARY_;
    
    MExpRetrievedMissionSummary_t summarizeRetrievedMission_(
        const vector<MExpRetrievedMission_t> retrievedMissions);
    MExpStateInfo_t highlightStates_(
        vector<MExpStateInfo_t> stateInfoList,
        MExpStateInfo_t lastHighlightedStateInfo,
        bool *highlighted);
    config_fsa_lists_t getFSALists_(void);
    string createPrefix_(void);
    MExpStateInfo_t highlightStates_(vector<MExpStateInfo_t> stateInfoList);
    int checkFeatureStatusFile_(void);
    int revertFile_(string filename);
    void nameDataFiles_(void);
    void runMlabMissionDesign_(void);
    void runMlabReplay_(void);
    void runCBRMExp_(void);
    void readCBRMExpResult_(void);
    void createTemporaryTaskAndGlobalFeatures_(void);
    void popupFeatureMenu_(bool askTaskType);
    void popupStatusWindow_(
        int rating,
        string missionSummary,
        bool prevBtnIsSensitive,
        bool nextBtnIsSensitive,
        bool loadBtnIsSensitive,
        bool cancelBtnIsSensitive,
        int currentMissionId);
    void updateStatusWindow_(
        int rating,
        string missionSummary,
        bool prevBtnIsSensitive,
        bool nextBtnIsSensitive,
        bool loadBtnIsSensitive,
        bool cancelBtnIsSensitive,
        int currentMissionId);
    void pendXCommandUpdateStatusWindow_(
        int rating,
        string missionSummary,
        bool prevBtnIsSensitive,
        bool nextBtnIsSensitive,
        bool loadBtnIsSensitive,
        bool cancelBtnIsSensitive,
        int currentMissionId);
    void selectMissionForStatusWindow_(void);
    void closeStatusWindow_(void);
    void changeStatusWindow_(int direction);
    void changeFeatureMenuValue_(XtPointer client_data);
    void changeGlobalFeatureMenuValue_(XtPointer client_data);
    void changeFeatureMenuTaskValue_(int taskId);
    void applyFeatureMenu_(void);
    void closeFeatureMenu_(void);
    void deleteCreatedDataFiles_(void);
    void addCreatedDataFileList_(string filename);
    void createPixmaps_(Pixel fg, Pixel bg, Widget parent);
    void askDiagnosis_(void);
    void startProcessThreadedXCommands_(void);
    void pendXCommandHighlightStates_(vector<MExpStateInfo_t> stateInfoList);
    void processThreadedXCommands_(void);
    void startCheckForkedMlabStatus_(int pid, bool eventLogIsON);
    void clearFeatureFile_(void);
    void askCNPMode_(bool simpleCheck);
    bool confirmViewAnyway_();
    bool initCBRClient_(void);
    bool buildMission_(void);
    bool repairMission_(void);
    bool loadRetrievedMissionChosenByUser_(
        vector<MExpRetrievedMission_t> retrievedMissions,
        string overlayFileName,
        int *selectedIndex);
    bool loadRetrievedMission_(
        vector<MExpRetrievedMission_t> retrievedMissions,
        string overlayFileName,
        int selectedIndex);
    bool hasAgentInMissionPlan_(string agentName);
    bool isSameStateInfo_(MExpStateInfo_t stateInfoA, MExpStateInfo_t stateInfoB);
    bool checkForkedProcessStatus_(int pid);
    bool saveAbstractedMissionPlanToFile_(Symbol * agent, bool *hasStartSubMission);

    // Callbacks
    static XtTimerCallbackProc cbProcessThreadedXCommands_(XtPointer clientData);
    static XtTimerCallbackProc cbCheckForkedMlabStatus_(XtPointer clientData);
    static void cbFeatureMenuTaskValue_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbChangeGlobalFeatureMenuValue_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbFeatureMenuValue_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbApplyFeatureMenu_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbCloseFeatureMenu_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbPrevBtnStatusWindow_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbNextBtnStatusWindow_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbLoadBtnStatusWindow_(Widget w, XtPointer client_data, XtPointer callback_data);
    static void cbCancelBtnStatusWindow_(Widget w, XtPointer client_data, XtPointer callback_data);

public:
    MissionExpert(
        Display *display,
        Widget parent,
        XtAppContext app,
        const symbol_table<rc_chain> &rcTable);
    ~MissionExpert(void);
    string getOverlayFileName(void);
    string getFeatureFileString(void);
    vector<MExpLogfileInfo_t> getLogfileInfoList(void);
    int assignedRobotID(int index);
    int getCNPMode(void);
    void initialize(void);
    void confirmSuccessfulMission(void);
    void runMissionExpert(int mode);
    void resumeMissionExpert_(void);
    void saveOverlayFileName(string overlayFileName);
    void saveSelectionToCBRLibrary(Symbol * agent);
    void madeChange(void);
    void saveLogfileInfo(MExpLogfileInfo_t logfileInfo);
    void clearLogfileInfoList(void);
    void highlightStates(vector<MExpStateInfo_t> stateInfoList);
    void saveSpecifiedFeatureFileName(string featureFileName);
    bool askUseMissionExpert(void);
    bool isDisabled(void);
    bool isEnabled(void);
    bool repairDisabled(void);
    bool shouldSaveMissionToCBRLibrary(void);
    bool shouldAskFeedback(void);
};

inline int MissionExpert::getCNPMode(void)
{
    return cnpMode_;
}

inline bool MissionExpert::isDisabled(void)
{
    return disableMExp_;
}

inline bool MissionExpert::isEnabled(void)
{
    return (!disableMExp_);
}

inline bool MissionExpert::repairDisabled(void)
{
    return disableRepair_;
}

inline bool MissionExpert::shouldSaveMissionToCBRLibrary(void)
{
    return shouldSaveMissionToCBRLibrary_;
}

inline void MissionExpert::selectMissionForStatusWindow_(void)
{
    statusWindows_.selectedMissionId =  statusWindows_.currentMissionId;
}

inline bool MissionExpert::isSameStateInfo_(
    MExpStateInfo_t stateInfoA,
    MExpStateInfo_t stateInfoB)
{
    return ((stateInfoA.fsaName == stateInfoB.fsaName) &&
            (stateInfoA.stateName == stateInfoB.stateName));
}

extern MissionExpert *gMExp;

#endif

/**********************************************************************
 * $Log: mission_expert.h,v $
 * Revision 1.1.1.1  2006/07/12 13:37:54  endo
 * MissionLab 7.0
 *
 * Revision 1.8  2006/05/15 01:23:28  endo
 * gcc-3.4 upgrade
 *
 * Revision 1.7  2006/02/14 02:27:18  endo
 * gAutomaticExecution flag and its capability added.
 *
 * Revision 1.6  2006/01/30 02:47:28  endo
 * AO-FNC CBR-CNP Type-I check-in.
 *
 * Revision 1.5  2005/08/09 19:12:44  endo
 * Things improved for the August demo.
 *
 * Revision 1.4  2005/07/31 03:39:43  endo
 * Robot ID specified by CNP can be now recognized by mlab during execution.
 *
 * Revision 1.3  2005/07/16 08:47:49  endo
 * CBR-CNP integration
 *
 * Revision 1.2  2005/02/07 22:25:27  endo
 * Mods for usability-2004
 *
 * Revision 1.1.1.1  2005/02/06 22:59:35  endo
 * AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
 *
 * Revision 1.3  2003/04/06 12:25:34  endo
 * gcc 3.1.1
 *
 * Revision 1.2  2003/04/06 08:50:37  endo
 * Updated for CBR Wizard Prototype II.
 *
 * Revision 1.1  2002/01/12 23:09:58  endo
 * Initial revision
 *
 **********************************************************************/
